"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const platform_1 = require("./platform");
const config_1 = __importDefault(require("./config"));
const Utils_1 = __importDefault(require("./Utils"));
const types_1 = require("./types");
class ProxyService {
    sdkInstance = {
        config: config_1.default,
        session: null,
    };
    requestsNumber = 0;
    fetchImpl = platform_1.fetchImpl;
    currentUserId;
    abortControllersMap = {};
    requestMethod = types_1.ProxyMethod;
    responseType = types_1.ProxyType;
    setSession(session) {
        this.sdkInstance.session = session;
        if (session && session.user_id) {
            this.setCurrentUserId(session.user_id);
        }
    }
    getSession() {
        return this.sdkInstance.session;
    }
    setCurrentUserId(userId) {
        this.currentUserId = userId || undefined;
    }
    getCurrentUserId() {
        return this.currentUserId;
    }
    logRequest(params, requestId) {
        Utils_1.default.DLog(`[Request][${requestId}]`, `${params.type || 'GET'} ${params.url}`, params);
    }
    logResponse(response, requestId) {
        Utils_1.default.DLog(`[Response][${requestId}]`, response);
    }
    buildRequestAndURL(params) {
        const isGetOrHeadType = !params.type || params.type === 'GET' || params.type === 'HEAD';
        const isPostOrPutType = params.type ? params.type === 'POST' || params.type === 'PUT' : false;
        const token = this.sdkInstance && this.sdkInstance.session && this.sdkInstance.session.token;
        const isInternalRequest = params.url.indexOf('s3.amazonaws.com') === -1;
        const isMultipartFormData = params.contentType === false;
        const authKey = params.authKey;
        let requestBody;
        let requestURL = params.url;
        const requestObject = {};
        requestObject.method = params.type || 'GET';
        if (params.data) {
            requestBody = this.buildRequestBody(params, isMultipartFormData, isPostOrPutType);
            if (isGetOrHeadType) {
                requestURL += `?${requestBody}`;
            }
            else {
                requestObject.body = requestBody;
            }
        }
        if (!isMultipartFormData) {
            requestObject.headers = {
                'Content-Type': isPostOrPutType
                    ? 'application/json;charset=utf-8'
                    : 'application/x-www-form-urlencoded; charset=UTF-8',
            };
        }
        if (isInternalRequest) {
            if (!requestObject.headers) {
                requestObject.headers = {};
            }
            requestObject.headers['CB-SDK'] = `JS ${config_1.default.version} - Client`;
            if (token) {
                requestObject.headers['CB-Token'] = token;
            }
            else if (authKey) {
                requestObject.headers['CB-AuthKey'] = authKey;
            }
        }
        if (config_1.default.timeout) {
            requestObject.timeout = config_1.default.timeout;
        }
        return [requestObject, requestURL];
    }
    buildRequestBody(params, isMultipartFormData, isPostOrPutType) {
        const data = params.data;
        const useArrayQuery = params.useArrayQuery;
        let dataObject;
        if (isMultipartFormData) {
            dataObject = new platform_1.formDataImpl();
            Object.keys(data).forEach(function (item) {
                if (params.fileToCustomObject && item === 'file') {
                    dataObject.append(item, data[item].data, data[item].name);
                }
                else {
                    dataObject.append(item, params.data[item]);
                }
            });
        }
        else if (isPostOrPutType) {
            dataObject = JSON.stringify(data);
        }
        else {
            dataObject = this.serializeQueryParams(data, '', useArrayQuery, 0);
        }
        return dataObject;
    }
    serializeQueryParams(obj, prefix, useArrayQuery, level = 0) {
        const parts = [];
        for (let propName in obj) {
            let propQueryName = this.encodeURIComponent(propName);
            if (Utils_1.default.isArray(obj)) {
                propQueryName = '';
            }
            const key = prefix ? prefix + `[${propQueryName}]` : propQueryName;
            let value = obj[propName];
            const isArrayVal = Utils_1.default.isArray(value);
            if ((isArrayVal && (useArrayQuery || level === 0)) || Utils_1.default.isObject(value)) {
                parts.push(this.serializeQueryParams(value, key, useArrayQuery, ++level));
            }
            else {
                value = isArrayVal ? value.sort().join(',') : value;
                parts.push(`${key}=${this.encodeURIComponent(value)}`);
            }
        }
        return parts.sort().join('&');
    }
    encodeURIComponent(uriComponent) {
        return encodeURIComponent(uriComponent).replace(/[#$&+,/:;=?@\[\]]/g, (c) => `%${c.charCodeAt(0).toString(16)}`);
    }
    abortRequest(abortId) {
        if (this.abortControllersMap[abortId]) {
            const controllers = this.abortControllersMap[abortId].controllers || [];
            controllers.forEach((controller) => {
                controller.abort();
            });
        }
    }
    processSuccessfulOrFailedRequest(abort_id) {
        if (!abort_id || !this.abortControllersMap[abort_id]) {
            return;
        }
        const controllers = this.abortControllersMap[abort_id].controllers || [];
        if (!this.abortControllersMap[abort_id].doneRequestsCount) {
            this.abortControllersMap[abort_id].doneRequestsCount = 1;
        }
        else {
            this.abortControllersMap[abort_id].doneRequestsCount += 1;
        }
        const doneRequestsCount = this.abortControllersMap[abort_id].doneRequestsCount;
        if (doneRequestsCount === controllers.length) {
            delete this.abortControllersMap[abort_id];
        }
    }
    async ajax(params) {
        const requestId = ++this.requestsNumber;
        return new Promise((resolve, reject) => {
            this.logRequest(params, requestId);
            const [requestObject, requestURL] = this.buildRequestAndURL(params);
            const abort_id = params.abort_id;
            if (abort_id) {
                let index = 0;
                if (this.abortControllersMap[abort_id]) {
                    const controllers = this.abortControllersMap[abort_id].controllers || [];
                    this.abortControllersMap[abort_id].controllers.push(new AbortController());
                    index = controllers.length - 1;
                }
                else {
                    this.abortControllersMap[abort_id] = {
                        controllers: [new AbortController()],
                    };
                }
                const signal = this.abortControllersMap[abort_id].controllers[index].signal;
                Object.assign(requestObject, { signal });
            }
            let response;
            // The Promise returned from fetch() won’t reject on HTTP error
            // status even if the response is an HTTP 404 or 500.
            // Instead, it will resolve normally (with ok status set to false),
            // and it will only reject on network failure or if anything prevented the request from completing.
            (0, platform_1.fetchImpl)(requestURL, requestObject)
                .then((resp) => {
                response = resp;
                const dataType = params.dataType || types_1.ProxyType.JSON;
                return dataType === types_1.ProxyType.TEXT ? response.text() : response.json();
            })
                .then((body) => {
                this.processSuccessfulOrFailedRequest(abort_id);
                if (!response.ok) {
                    this.processAjaxError(response, body, null, reject, resolve, params, requestId);
                }
                else {
                    this.processAjaxResponse(body, resolve, requestId);
                }
            })
                .catch((error) => {
                this.processSuccessfulOrFailedRequest(abort_id);
                this.processAjaxError(response, ' ', error, reject, resolve, params, requestId);
            });
        });
    }
    processAjaxResponse(body, resolve, requestId) {
        const responseBody = body && body !== ' ' ? body : 'empty body';
        this.logResponse(responseBody, requestId);
        resolve(body);
    }
    processAjaxError(response, body, error, reject, resolve, params, requestId) {
        if (!response && error && !error.code) {
            reject(error);
            return;
        }
        const responseBody = body || error || body.errors;
        const statusCode = response?.status;
        const errorObject = {
            code: (response && statusCode) || (error && error.code),
            info: (body && typeof body === 'string' && body !== ' ' ? JSON.parse(body) : body) || (error && error.errno),
        };
        this.logResponse(responseBody, requestId);
        if (response?.url.indexOf(config_1.default.urls.session) === -1) {
            if (this.isExpiredSessionError(errorObject) && typeof config_1.default.on.sessionExpired === 'function') {
                this.handleExpiredSessionResponse(errorObject, null, reject, resolve, params);
            }
            else {
                reject(errorObject);
            }
        }
        else {
            reject(errorObject);
        }
    }
    handleExpiredSessionResponse(error, response, reject, resolve, params) {
        const handleResponse = () => {
            if (error) {
                reject(error);
            }
            else {
                resolve(response);
            }
        };
        const retryCallback = (session) => {
            if (session) {
                this.setSession(session);
                this.ajax(params).then(resolve).catch(reject);
            }
        };
        if (typeof config_1.default.on.sessionExpired === 'function') {
            config_1.default.on.sessionExpired(handleResponse, retryCallback);
        }
    }
    isExpiredSessionError(error) {
        try {
            return error && error.code === 401 && error.info?.errors?.base?.[0] === 'Required session does not exist';
        }
        catch {
            return false;
        }
    }
}
exports.default = ProxyService;
//# sourceMappingURL=Proxy.js.map