"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const config_1 = __importDefault(require("../config"));
const xmpp_1 = require("./xmpp");
const Utils_1 = __importDefault(require("../Utils"));
const Dialogs_1 = __importDefault(require("./Dialogs"));
const Messages_1 = __importDefault(require("./Messages"));
const Helpers_1 = __importDefault(require("./Helpers"));
const ChatUtils_1 = __importDefault(require("./ChatUtils"));
const StreamManagement_1 = __importDefault(require("./StreamManagement"));
const PrivacyList_1 = __importDefault(require("./PrivacyList"));
const MultiUserChat_1 = __importDefault(require("./MultiUserChat"));
const types_1 = require("../types");
class ChatService {
    proxy;
    dialog;
    message;
    xmppClient;
    helpers;
    privacylist;
    muc;
    streamManagement;
    webrtcSignalingProcessor;
    isConnected = false;
    isConnecting = false;
    isLogout = false;
    isReconnect = false;
    stanzasCallbacks = {};
    xmppClientListeners = new Map();
    connectPromise = null;
    earlyIncomingMessagesQueue = [];
    onChatStatusListener;
    onConnectionErrorListener;
    onDisconnectedListener;
    onReconnectListener;
    onMessageListener;
    onSystemMessageListener;
    onMessageErrorListener;
    onMessageTypingListener;
    onMessageUpdateListener;
    onMessageDeleteListener;
    onMessageReactionsListener;
    onSentMessageCallback;
    onDeliveredStatusListener;
    onReadStatusListener;
    onLastUserActivityListener;
    onJoinOccupant;
    onLeaveOccupant;
    onKickOccupant;
    constructor(proxy) {
        this.proxy = proxy;
        this.dialog = new Dialogs_1.default(proxy);
        this.message = new Messages_1.default(proxy);
        this.helpers = new Helpers_1.default();
        this.xmppClient = (0, xmpp_1.xmppClient)({
            service: config_1.default.chatProtocol.websocket,
            credentials: (auth, _mechanism) => auth({
                username: this.xmppClient.options.username || '',
                password: this.xmppClient.options.password || '',
            }),
        });
        if (config_1.default.chat.reconnect.enable) {
            this.xmppClient.reconnect.delay = config_1.default.chat.reconnect.timeInterval * 1000;
        }
        const options = {
            xmppClient: this.xmppClient,
            helpers: this.helpers,
            stanzasCallbacks: this.stanzasCallbacks,
        };
        this.muc = new MultiUserChat_1.default(options);
        this.privacylist = new PrivacyList_1.default(options);
        this.streamManagement = new StreamManagement_1.default(this);
    }
    get connectionStatus() {
        return this.xmppClient.status;
    }
    async connect(params) {
        this.connectPromise = new Promise((resolve, reject) => {
            Utils_1.default.DLog('[Chat]', 'Connect with parameter:', params);
            if (this.isConnecting) {
                Utils_1.default.DLog('[Chat]', 'Warning! Already in CONNECTING state');
                resolve(false);
                return;
            }
            if (this.isConnected) {
                Utils_1.default.DLog('[Chat]', 'Warning! Chat is already connected!');
                resolve(false);
                return;
            }
            this.isConnecting = true;
            this.isLogout = false;
            this.removeAllXMPPClientListeners();
            this.addXMPPClientListener('connect', () => {
                Utils_1.default.DLog('[Chat]', this.isReconnect ? 'RECONNECTING' : 'CONNECTING');
            });
            this.addXMPPClientListener('online', () => {
                Utils_1.default.DLog('[Chat]', 'ONLINE');
                this.postConnectActions();
                resolve(true);
                this.connectPromise = null;
            });
            this.addXMPPClientListener('offline', () => {
                Utils_1.default.DLog('[Chat]', 'OFFLINE');
            });
            this.addXMPPClientListener('disconnect', () => {
                Utils_1.default.DLog('[Chat]', 'DISCONNECTED');
                Utils_1.default.safeCallbackCall(this.onDisconnectedListener)();
                if (config_1.default.chat.reconnect.enable) {
                    this.isConnected = false;
                    this.isConnecting = false;
                }
                else {
                    this.disconnect();
                }
            });
            this.addXMPPClientListener('status', (status, value) => {
                Utils_1.default.DLog('[Chat]', `status - ${status}`, typeof value === 'object' ? JSON.stringify(value) : '');
                Utils_1.default.safeCallbackCall(this.onChatStatusListener)(status);
            });
            this.addXMPPClientListener('stanza', (stanza) => {
                // it can be a case,
                // when message came after xmpp auth but before resource binding,
                // and it can cause some crashes, e.g.
                // https://github.com/ConnectyCube/connectycube-js-sdk-releases/issues/28
                if (stanza.is('message') && !this.isConnected) {
                    this.earlyIncomingMessagesQueue.push(stanza);
                    Utils_1.default.DLog('[Chat]', "on 'stanza': enqueue incoming stanza (isConnected=false)");
                    return;
                }
                // after 'input' and 'element' (only if stanza, not nonza)
                if (stanza.is('presence')) {
                    this.onPresence(stanza);
                }
                else if (stanza.is('iq')) {
                    this.onIQ(stanza);
                }
                else if (stanza.is('message')) {
                    if (stanza.attrs.type === 'headline') {
                        this.onSystemMessage(stanza);
                    }
                    else if (stanza.attrs.type === 'error') {
                        this.onMessageError(stanza);
                    }
                    else {
                        this.onMessage(stanza);
                    }
                }
            });
            this.addXMPPClientListener('error', (error) => {
                Utils_1.default.DLog('[Chat]', 'ERROR:', error, {
                    isReconnect: this.isReconnect,
                    connectPromise: !!this.connectPromise,
                });
                if (this.connectPromise) {
                    if (!this.isReconnect) {
                        reject(error.name == 'SASLError' ? error.condition : error);
                        this.connectPromise = null;
                    }
                }
                else {
                    Utils_1.default.safeCallbackCall(this.onConnectionErrorListener)(error);
                }
            });
            this.addXMPPClientListener('output', (str) => {
                Utils_1.default.callTrafficUsageCallback('xmppDataWrite', { body: str });
                Utils_1.default.DLog('[Chat]', 'SENT:', str);
            });
            this.addXMPPClientListener('input', (str) => {
                Utils_1.default.callTrafficUsageCallback('xmppDataRead', { body: str });
                Utils_1.default.DLog('[Chat]', 'RECV:', str);
            });
            // save user connection data so they will be used when authenticate (above)
            this.xmppClient.options.username = ChatUtils_1.default.buildUserJidLocalPart(params.userId);
            this.xmppClient.options.password = params.password;
            // start connect
            this.xmppClient.start().catch((error) => {
                console.error('[Chat] xmppClient.start error', error);
                if (this.connectPromise) {
                    reject(error);
                    this.connectPromise = null;
                }
                else {
                    Utils_1.default.safeCallbackCall(this.onConnectionErrorListener)(error);
                }
            });
        });
        return this.connectPromise;
    }
    async ping() {
        return new Promise((resolve, reject) => {
            const id = ChatUtils_1.default.getUniqueId('ping');
            const iqStanza = ChatUtils_1.default.createIqStanza({ id, type: 'get', to: config_1.default.endpoints.chat });
            iqStanza.c('ping', { xmlns: 'urn:xmpp:ping' });
            this.stanzasCallbacks[id] = (stanza) => {
                const error = ChatUtils_1.default.getElement(stanza, 'error');
                if (error) {
                    reject(ChatUtils_1.default.buildErrorFromXMPPErrorStanza(error));
                }
                else {
                    resolve();
                }
            };
            this.xmppClient.sendOnline(iqStanza);
        });
    }
    async pingWithTimeout(timeout = 5000) {
        return Promise.race([
            this.ping(),
            new Promise((_, reject) => setTimeout(() => {
                const error = 'Chat ping() timed out';
                Utils_1.default.safeCallbackCall(this.onConnectionErrorListener)(error);
                reject(error);
            }, timeout)),
        ]);
    }
    send(jidOrUserId, message = {}) {
        const messageId = message.id ?? Utils_1.default.getBsonObjectId();
        const messageStanza = ChatUtils_1.default.createMessageStanza({
            from: this.helpers.userCurrentJid,
            to: this.helpers.jidOrUserId(jidOrUserId),
            type: message.type ?? this.helpers.typeChat(jidOrUserId),
            id: messageId,
        });
        if (!message.id) {
            message.id = messageId;
        }
        if (message.body) {
            messageStanza.c('body').t(message.body).up();
        }
        if (message.markable) {
            messageStanza.c('markable', { xmlns: 'urn:xmpp:chat-markers:0' }).up();
        }
        if (message.extension) {
            messageStanza.c('extraParams', { xmlns: 'jabber:client' });
            ChatUtils_1.default.assignExtraParamsToXml(messageStanza, message.extension);
        }
        if (this.streamManagement.supported) {
            this.streamManagement.sendAndCount(messageStanza, message);
        }
        else {
            this.xmppClient.sendOnline(messageStanza);
        }
        return messageId;
    }
    sendSystemMessage(jidOrUserId, message) {
        const messageId = message.id ?? Utils_1.default.getBsonObjectId();
        const messageStanza = ChatUtils_1.default.createMessageStanza({
            type: 'headline',
            id: messageId,
            to: this.helpers.jidOrUserId(jidOrUserId),
        });
        if (message.body) {
            messageStanza.c('body').t(message.body).up();
        }
        if (message.extension) {
            messageStanza.c('extraParams', { xmlns: 'jabber:client' }).c('moduleIdentifier').t('SystemNotifications').up();
            ChatUtils_1.default.assignExtraParamsToXml(messageStanza, message.extension);
        }
        this.xmppClient.sendOnline(messageStanza);
        return messageId;
    }
    sendIsTypingStatus(jidOrUserId) {
        const messageStanza = ChatUtils_1.default.createMessageStanza({
            from: this.helpers.userCurrentJid,
            to: this.helpers.jidOrUserId(jidOrUserId),
            type: this.helpers.typeChat(jidOrUserId),
        });
        messageStanza.c('composing', { xmlns: 'http://jabber.org/protocol/chatstates' });
        this.xmppClient.sendOnline(messageStanza);
    }
    sendIsStopTypingStatus(jidOrUserId) {
        const messageStanza = ChatUtils_1.default.createMessageStanza({
            from: this.helpers.userCurrentJid,
            to: this.helpers.jidOrUserId(jidOrUserId),
            type: this.helpers.typeChat(jidOrUserId),
        });
        messageStanza.c('paused', { xmlns: 'http://jabber.org/protocol/chatstates' });
        this.xmppClient.sendOnline(messageStanza);
    }
    sendDeliveredStatus(params) {
        const messageStanza = ChatUtils_1.default.createMessageStanza({
            type: 'chat',
            from: this.helpers.userCurrentJid,
            id: Utils_1.default.getBsonObjectId(),
            to: this.helpers.jidOrUserId(params.userId),
        });
        messageStanza.c('received', { xmlns: 'urn:xmpp:chat-markers:0', id: params.messageId }).up();
        messageStanza.c('extraParams', { xmlns: 'jabber:client' }).c('dialog_id').t(params.dialogId);
        this.xmppClient.sendOnline(messageStanza);
    }
    sendReadStatus(params) {
        const messageStanza = ChatUtils_1.default.createMessageStanza({
            type: 'chat',
            from: this.helpers.userCurrentJid,
            to: this.helpers.jidOrUserId(params.userId),
            id: Utils_1.default.getBsonObjectId(),
        });
        messageStanza.c('displayed', { xmlns: 'urn:xmpp:chat-markers:0', id: params.messageId }).up();
        messageStanza.c('extraParams', { xmlns: 'jabber:client' }).c('dialog_id').t(params.dialogId);
        this.xmppClient.sendOnline(messageStanza);
    }
    editMessage(params) {
        const messageStanza = ChatUtils_1.default.createMessageStanza({
            from: this.helpers.userCurrentJid,
            to: this.helpers.jidOrUserId(params.to),
            type: this.helpers.typeChat(params.to),
            id: Utils_1.default.getBsonObjectId(),
        });
        messageStanza.c('body').t(params.body).up();
        messageStanza
            .c('replace', {
            xmlns: 'urn:xmpp:message-correct:0',
            id: params.originMessageId,
            last: params.last ? 'true' : 'false',
        })
            .up();
        messageStanza.c('extraParams', { xmlns: 'jabber:client' }).c('dialog_id').t(params.dialogId);
        if (params.extension) {
            ChatUtils_1.default.assignExtraParamsToXml(messageStanza, params.extension);
        }
        this.xmppClient.sendOnline(messageStanza);
    }
    deleteMessage(params) {
        const messageStanza = ChatUtils_1.default.createMessageStanza({
            from: this.helpers.userCurrentJid,
            to: this.helpers.jidOrUserId(params.to),
            type: this.helpers.typeChat(params.to),
            id: Utils_1.default.getBsonObjectId(),
        });
        messageStanza.c('remove', { xmlns: 'urn:xmpp:message-delete:0', id: params.messageId }).up();
        messageStanza.c('extraParams', { xmlns: 'jabber:client' }).c('dialog_id').t(params.dialogId);
        this.xmppClient.sendOnline(messageStanza);
    }
    getLastUserActivity(jidOrUserId) {
        return new Promise((resolve, reject) => {
            const id = ChatUtils_1.default.getUniqueId('lastActivity');
            const iqStanza = ChatUtils_1.default.createIqStanza({
                type: 'get',
                from: this.helpers.userCurrentJid,
                to: this.helpers.jidOrUserId(jidOrUserId),
                id: id,
            });
            iqStanza.c('query', { xmlns: 'jabber:iq:last' });
            this.stanzasCallbacks[id] = (stanza) => {
                if (ChatUtils_1.default.getElement(stanza, 'error')) {
                    reject(ChatUtils_1.default.buildErrorFromXMPPErrorStanza(stanza));
                }
                else {
                    resolve(this.onLastActivityStanza(stanza));
                }
            };
            this.xmppClient.sendOnline(iqStanza);
        });
    }
    onLastActivityStanza(stanza) {
        const from = ChatUtils_1.default.getAttr(stanza, 'from');
        const userId = this.helpers.getUserIdFromJID(from);
        const query = ChatUtils_1.default.getElement(stanza, 'query');
        const seconds = query ? parseInt(ChatUtils_1.default.getAttr(query, 'seconds')) : 0;
        Utils_1.default.safeCallbackCall(this.onLastUserActivityListener)(userId, seconds);
        return { userId, seconds };
    }
    markActive() {
        const iqStanza = ChatUtils_1.default.createIqStanza({
            id: this.helpers.getUniqueId('markActive'),
            type: 'set',
        });
        iqStanza.c('mobile', {
            xmlns: 'http://tigase.org/protocol/mobile#v2',
            enable: 'false',
        });
        this.xmppClient.sendOnline(iqStanza);
    }
    markInactive() {
        const iqStanza = ChatUtils_1.default.createIqStanza({
            id: this.helpers.getUniqueId('markActive'),
            type: 'set',
        });
        iqStanza.c('mobile', {
            xmlns: 'http://tigase.org/protocol/mobile#v2',
            enable: 'true',
        });
        this.xmppClient.sendOnline(iqStanza);
    }
    async disconnect() {
        Utils_1.default.DLog('[Chat]', 'disconnect');
        if (this.isLogout) {
            Utils_1.default.DLog('[Chat]', 'Warning! Chat is already disconnected!');
            return true;
        }
        this.isLogout = true;
        this.isReconnect = false;
        this.isConnected = false;
        this.isConnecting = false;
        this.muc.joinedRooms = {};
        this.helpers.userCurrentJid = '';
        this.streamManagement.stop();
        try {
            await this.xmppClient.stop();
            return true;
        }
        catch {
            return false;
        }
    }
    terminate() {
        Utils_1.default.DLog('[Chat]', 'terminated');
        this.isConnected = false;
        this.isConnecting = false;
        this.streamManagement.stop();
        this.xmppClient.socket.end();
    }
    async search(params) {
        const query = Object.assign({}, params);
        if (query.start_date) {
            query.start_date = new Date(query.start_date).toISOString();
        }
        if (query.end_date) {
            query.end_date = new Date(query.end_date).toISOString();
        }
        if (Utils_1.default.isArray(query.chat_dialog_ids)) {
            query.chat_dialog_ids = query.chat_dialog_ids.join(',');
        }
        const ajaxParams = {
            type: types_1.ProxyMethod.GET,
            url: Utils_1.default.getUrl(`${config_1.default.urls.chat}/search`),
            data: query,
        };
        return this.proxy.ajax(ajaxParams);
    }
    onMessage(rawStanza) {
        const forwardedStanza = ChatUtils_1.default.getElementTreePath(rawStanza, ['sent', 'forwarded', 'message']);
        const stanza = forwardedStanza || rawStanza;
        const from = ChatUtils_1.default.getAttr(stanza, 'from');
        const type = ChatUtils_1.default.getAttr(stanza, 'type');
        const invite = ChatUtils_1.default.getElement(stanza, 'invite');
        const resource = this.xmppClient.jid?.getResource() || '';
        const isChat = type === 'chat';
        // ignore private message from the same resource or invite messages from MUC
        if ((isChat && from.includes(resource)) || invite) {
            return;
        }
        const messageId = ChatUtils_1.default.getAttr(stanza, 'id');
        const markable = ChatUtils_1.default.getElement(stanza, 'markable');
        const delivered = ChatUtils_1.default.getElement(stanza, 'received');
        const read = ChatUtils_1.default.getElement(stanza, 'displayed');
        const replaceSubElement = ChatUtils_1.default.getElement(stanza, 'replace');
        const reactionsSubElement = ChatUtils_1.default.getElement(stanza, 'reactions');
        const removeSubElement = ChatUtils_1.default.getElement(stanza, 'remove');
        const composing = Boolean(ChatUtils_1.default.getElement(stanza, 'composing'));
        const paused = ChatUtils_1.default.getElement(stanza, 'paused');
        const delay = ChatUtils_1.default.getElement(stanza, 'delay');
        const extraParams = ChatUtils_1.default.getElement(stanza, 'extraParams');
        const body = ChatUtils_1.default.getElementText(stanza, 'body');
        const isForwarded = Boolean(forwardedStanza);
        const isGroup = type === 'groupchat';
        const recipient = isChat ? ChatUtils_1.default.getAttr(stanza, 'to') : null;
        const recipientId = recipient ? this.helpers.getUserIdFromJID(recipient) : null;
        const ext = extraParams ? ChatUtils_1.default.parseExtraParams(extraParams) : null;
        const extension = ext ? ext.extension : null;
        const dialogId = isGroup ? this.helpers.getDialogIdFromJID(from) : ext?.dialogId ?? null;
        const userId = isGroup ? this.helpers.getIdFromResource(from) : this.helpers.getUserIdFromJID(from);
        const currentUserJid = this.xmppClient.jid?.toString();
        const currentUserId = currentUserJid ? this.helpers.getUserIdFromJID(currentUserJid) : 0;
        // typing statuses
        if (composing || paused) {
            if (isChat || isGroup || !delay) {
                Utils_1.default.safeCallbackCall(this.onMessageTypingListener)(composing, userId, dialogId);
            }
            return;
        }
        // edit message
        if (replaceSubElement) {
            Utils_1.default.safeCallbackCall(this.onMessageUpdateListener)(ChatUtils_1.default.getAttr(replaceSubElement, 'id'), ChatUtils_1.default.getAttr(replaceSubElement, 'last') === 'true', body, dialogId, userId, extension);
            return;
        }
        // reactions
        if (reactionsSubElement) {
            if (isForwarded && isGroup) {
                return;
            }
            const messageId = ChatUtils_1.default.getAttr(reactionsSubElement, 'message_id');
            const userId = parseInt(ChatUtils_1.default.getAttr(reactionsSubElement, 'user_id'));
            const { add, remove } = ChatUtils_1.default.parseReactions(reactionsSubElement);
            Utils_1.default.safeCallbackCall(this.onMessageReactionsListener)(messageId, userId, dialogId, add, remove);
            return;
        }
        // delete message
        if (removeSubElement) {
            const messageId = ChatUtils_1.default.getAttr(removeSubElement, 'id');
            Utils_1.default.safeCallbackCall(this.onMessageDeleteListener)(messageId, dialogId, userId);
            return;
        }
        // delivered / read statuses
        if (delivered || read) {
            if (delivered && isChat) {
                const id = ChatUtils_1.default.getAttr(delivered, 'id');
                Utils_1.default.safeCallbackCall(this.onDeliveredStatusListener)(id, dialogId, userId);
            }
            if (read && isChat) {
                const id = ChatUtils_1.default.getAttr(read, 'id');
                Utils_1.default.safeCallbackCall(this.onReadStatusListener)(id, dialogId, userId);
            }
            return;
        }
        // auto-send 'received' status (ignore messages from yourself)
        if (markable && dialogId && userId && userId !== currentUserId) {
            this.sendDeliveredStatus({ messageId, dialogId, userId });
        }
        const message = {
            id: messageId,
            dialog_id: dialogId,
            recipient_id: recipientId,
            is_forwarded: isForwarded,
            type,
            body,
            extension,
            delay,
        };
        if (markable) {
            message.markable = 1;
        }
        if (isChat || isGroup) {
            Utils_1.default.safeCallbackCall(this.onMessageListener)(userId, message);
        }
    }
    onPresence(stanza) {
        const from = ChatUtils_1.default.getAttr(stanza, 'from');
        const id = ChatUtils_1.default.getAttr(stanza, 'id');
        const type = ChatUtils_1.default.getAttr(stanza, 'type');
        const currentUserJid = this.xmppClient.jid?.toString();
        const currentUserId = this.helpers.getUserIdFromJID(currentUserJid);
        const x = ChatUtils_1.default.getElement(stanza, 'x');
        const xmlns = x ? ChatUtils_1.default.getAttr(x, 'xmlns') : null;
        const status = x ? ChatUtils_1.default.getElement(x, 'xmlns') : null;
        // const statusCode = status ? ChatUtils.getElement(status, 'code') : null;
        const statusCode = status ? ChatUtils_1.default.getElementText(status, 'code') : null;
        // MUC presences
        if (xmlns && xmlns.startsWith('http://jabber.org/protocol/muc')) {
            // Error
            if (type === 'error') {
                // JOIN to dialog error
                if (id.endsWith(':join')) {
                    Utils_1.default.safeCallbackCall(this.stanzasCallbacks[id])(stanza);
                }
                return;
            }
            const dialogId = this.helpers.getDialogIdFromJID(from);
            const mucUserId = this.helpers.getUserIdFromRoomJid(from);
            // self presence
            if (status) {
                // KICK from dialog event
                if (statusCode === '301') {
                    const item = ChatUtils_1.default.getElement(stanza, 'item');
                    const actorElement = item ? ChatUtils_1.default.getElement(item, 'actor') : null;
                    const initiatorUserJid = actorElement ? ChatUtils_1.default.getAttr(actorElement, 'jid') : null;
                    const initiatorId = this.helpers.getUserIdFromJID(initiatorUserJid);
                    const roomJid = this.helpers.getRoomJidFromRoomFullJid(from);
                    Utils_1.default.safeCallbackCall(this.onKickOccupant)(dialogId, initiatorId);
                    if (roomJid) {
                        delete this.muc.joinedRooms[roomJid];
                    }
                    return;
                }
                else {
                    if (type === 'unavailable') {
                        // LEAVE response
                        if (status && statusCode === '110') {
                            Utils_1.default.safeCallbackCall(this.stanzasCallbacks['muc:leave'])(null);
                        }
                        return;
                    }
                    // JOIN response
                    if (id.endsWith(':join') && status && statusCode === '110') {
                        this.stanzasCallbacks[id]?.(stanza);
                        return;
                    }
                }
                // Occupants JOIN/LEAVE events
            }
            else if (mucUserId != currentUserId) {
                const listenerName = type === 'unavailable' ? 'onLeaveOccupant' : 'onJoinOccupant';
                Utils_1.default.safeCallbackCall(this[listenerName])(dialogId, mucUserId);
                return;
            }
        }
    }
    onIQ(stanza) {
        const stanzaId = ChatUtils_1.default.getAttr(stanza, 'id');
        if (this.stanzasCallbacks[stanzaId]) {
            Utils_1.default.safeCallbackCall(this.stanzasCallbacks[stanzaId])(stanza);
            delete this.stanzasCallbacks[stanzaId];
        }
        else {
            const from = ChatUtils_1.default.getAttr(stanza, 'from');
            const query = ChatUtils_1.default.getElement(stanza, 'query');
            if (from && query) {
                this.onLastActivityStanza(stanza);
            }
        }
    }
    onSystemMessage(rawStanza) {
        const forwardedStanza = ChatUtils_1.default.getElementTreePath(rawStanza, ['sent', 'forwarded', 'message']);
        const stanza = forwardedStanza || rawStanza;
        const from = ChatUtils_1.default.getAttr(stanza, 'from');
        const messageId = ChatUtils_1.default.getAttr(stanza, 'id');
        const extraParams = ChatUtils_1.default.getElement(stanza, 'extraParams');
        const userId = this.helpers.getUserIdFromJID(from);
        const delay = ChatUtils_1.default.getElement(stanza, 'delay');
        const moduleIdentifier = extraParams ? ChatUtils_1.default.getElementText(extraParams, 'moduleIdentifier') : null;
        const body = ChatUtils_1.default.getElementText(stanza, 'body');
        const ext = extraParams ? ChatUtils_1.default.parseExtraParams(extraParams) : null;
        const extension = ext ? ext.extension : null;
        switch (moduleIdentifier) {
            case 'SystemNotifications':
                Utils_1.default.safeCallbackCall(this.onSystemMessageListener)({ id: messageId, userId, body, extension });
                break;
            case 'WebRTCVideoChat':
                if (this.webrtcSignalingProcessor && !delay) {
                    Utils_1.default.safeCallbackCall(this.webrtcSignalingProcessor.onMessage)(userId, extraParams);
                }
                break;
            default:
                break;
        }
    }
    onMessageError(stanza) {
        // <error code="503" type="cancel">
        //   <service-unavailable xmlns="urn:ietf:params:xml:ns:xmpp-stanzas"/>
        //   <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" xml:lang="en">Service not available.</text>
        // </error>
        const messageId = ChatUtils_1.default.getAttr(stanza, 'id');
        const error = ChatUtils_1.default.buildErrorFromXMPPErrorStanza(stanza);
        Utils_1.default.safeCallbackCall(this.onMessageErrorListener)(messageId, error);
    }
    postConnectActions() {
        Utils_1.default.DLog('[Chat]', this.isReconnect ? 'RECONNECTED' : 'CONNECTED');
        const presence = ChatUtils_1.default.createPresenceStanza();
        this.streamManagement.start();
        this.helpers.userCurrentJid = this.xmppClient.jid?.toString() ?? null;
        this.isConnected = true;
        this.isConnecting = false;
        this.enableCarbons();
        this.xmppClient.sendOnline(presence); // initial presence
        if (this.isReconnect) {
            // reconnect
            Utils_1.default.safeCallbackCall(this.onReconnectListener)();
        }
        else {
            this.isReconnect = true;
        }
        if (this.earlyIncomingMessagesQueue.length > 0) {
            Utils_1.default.DLog('[Chat]', `Flush 'earlyIncomingMessagesQueue' (length=${this.earlyIncomingMessagesQueue.length})`);
            const stanzasCallback = this.xmppClientListeners.get('stanza');
            this.earlyIncomingMessagesQueue.forEach((stanza) => {
                stanzasCallback?.(stanza);
            });
            this.earlyIncomingMessagesQueue = [];
        }
    }
    enableCarbons() {
        const iqStanza = ChatUtils_1.default.createIqStanza({
            type: 'set',
            from: this.helpers.userCurrentJid,
            id: ChatUtils_1.default.getUniqueId('enableCarbons'),
        });
        iqStanza.c('enable', { xmlns: 'urn:xmpp:carbons:2' });
        this.xmppClient.sendOnline(iqStanza);
    }
    setSubscriptionToUserLastActivity(jidOrUserId, enable) {
        const iqStanza = ChatUtils_1.default.createIqStanza({
            id: this.helpers.getUniqueId('statusStreaming'),
            type: 'set',
        });
        iqStanza.c('subscribe', {
            xmlns: 'https://connectycube.com/protocol/status_streaming',
            user_jid: this.helpers.jidOrUserId(jidOrUserId),
            enable,
        });
        this.xmppClient.sendOnline(iqStanza);
    }
    subscribeToUserLastActivityStatus(jidOrUserId) {
        this.setSubscriptionToUserLastActivity(jidOrUserId, true);
    }
    unsubscribeFromUserLastActivityStatus(jidOrUserId) {
        this.setSubscriptionToUserLastActivity(jidOrUserId, false);
    }
    addXMPPClientListener(name, listener) {
        this.xmppClient.on(name, listener);
        this.xmppClientListeners.set(name, listener);
    }
    removeXMPPClientListener(name) {
        const listener = this.xmppClientListeners.get(name);
        if (listener) {
            this.xmppClient.removeListener(name, listener);
        }
        this.xmppClientListeners.delete(name);
    }
    removeAllXMPPClientListeners() {
        this.xmppClientListeners.forEach((_, name) => this.removeXMPPClientListener(name));
    }
    getListenerByName(name) {
        switch (name) {
            case types_1.ChatEvent.STATUS:
                return 'onChatStatusListener';
            case types_1.ChatEvent.ERROR:
                return 'onConnectionErrorListener';
            case types_1.ChatEvent.DISCONNECTED:
                return 'onDisconnectedListener';
            case types_1.ChatEvent.RECONNECTED:
                return 'onReconnectListener';
            case types_1.ChatEvent.MESSAGE:
                return 'onMessageListener';
            case types_1.ChatEvent.SYSTEM_MESSAGE:
                return 'onSystemMessageListener';
            case types_1.ChatEvent.ERROR_MESSAGE:
                return 'onMessageErrorListener';
            case types_1.ChatEvent.TYPING_MESSAGE:
                return 'onMessageTypingListener';
            case types_1.ChatEvent.UPDATE_MESSAGE:
                return 'onMessageUpdateListener';
            case types_1.ChatEvent.DELETE_MESSAGE:
                return 'onMessageDeleteListener';
            case types_1.ChatEvent.REACTIONS_MESSAGE:
                return 'onMessageReactionsListener';
            case types_1.ChatEvent.DELIVERED_MESSAGE:
                return 'onDeliveredStatusListener';
            case types_1.ChatEvent.READ_MESSAGE:
                return 'onReadStatusListener';
            case types_1.ChatEvent.SENT_MESSAGE:
                return 'onSentMessageCallback';
            case types_1.ChatEvent.USER_LAST_ACTIVITY:
                return 'onLastUserActivityListener';
            case types_1.ChatEvent.JOIN:
                return 'onJoinOccupant';
            case types_1.ChatEvent.LEAVE:
                return 'onLeaveOccupant';
            case types_1.ChatEvent.KICK:
                return 'onKickOccupant';
            default:
                return null;
        }
    }
    addListener(name, listener) {
        const listenerName = this.getListenerByName(name);
        if (listenerName) {
            this[listenerName] = listener;
        }
        return this.removeListener.bind(this, name);
    }
    removeListener(name) {
        const listenerName = this.getListenerByName(name);
        if (listenerName) {
            this[listenerName] = undefined;
        }
    }
    removeAllListeners() {
        Object.keys(this).forEach((key) => {
            if (key.startsWith('on') && key.endsWith('Listener') && typeof this[key] === 'function') {
                this[key] = undefined;
            }
        });
    }
}
exports.default = ChatService;
//# sourceMappingURL=Chat.js.map