"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const config_1 = __importDefault(require("../config"));
const xmpp_1 = require("./xmpp");
class ChatUtils {
    static buildUserJid(params = {}) {
        const { userId, resource, jid } = params;
        const appId = config_1.default.creds.appId;
        const url = config_1.default.endpoints.chat;
        return userId ? `${userId}-${appId}@${url}${resource ? `/${resource}` : ''}` : jid;
    }
    static buildUserJidLocalPart(userId) {
        return `${userId}-${config_1.default.creds.appId}`;
    }
    static createMessageStanza(attrs) {
        return (0, xmpp_1.xml)('message', attrs);
    }
    static createIqStanza(attrs) {
        return (0, xmpp_1.xml)('iq', attrs);
    }
    static createPresenceStanza(attrs) {
        return (0, xmpp_1.xml)('presence', attrs);
    }
    static createNonza(name, attrs) {
        return (0, xmpp_1.xml)(name, attrs);
    }
    static getAttr(element, name) {
        return element?.getAttr(name) || element?.attrs[name] || null;
    }
    static getElement(element, name) {
        return element?.getChild(name);
    }
    static getName(element) {
        return element?.getName() || null;
    }
    static getText(element) {
        return element?.getText() || null;
    }
    static getChildElements(element) {
        return element?.getChildElements() || [];
    }
    static isErrorStanza(element) {
        return !!element?.getChild('error');
    }
    static getAllElements(element, name) {
        return element?.getChildren(name) || [];
    }
    static getElementText(element, name) {
        return element?.getChildText(name) || null;
    }
    static getElementTreePath(element, elementsPath) {
        return elementsPath.reduce((prevElement, name) => prevElement ? ChatUtils.getElement(prevElement, name) : undefined, element);
    }
    static assignObjectToXml(element, obj, name) {
        element = element.c(name);
        Object.keys(obj).forEach((key) => {
            if (typeof obj[key] === 'object') {
                ChatUtils.assignObjectToXml(element, obj[key], key);
            }
            else {
                element = element.c(key).t(obj[key]).up();
            }
        });
        element = element.up();
        return element;
    }
    static assignExtraParamsToXml(element, extension) {
        let extraParams = ChatUtils.getElement(element, 'extraParams');
        Object.keys(extension).forEach((key) => {
            if (key === 'attachments') {
                extension[key].forEach((attachment) => {
                    extraParams.c('attachment', attachment).up();
                });
            }
            else if (typeof extension[key] === 'object') {
                ChatUtils.assignObjectToXml(extraParams, extension[key], key);
            }
            else {
                extraParams.c(key).t(extension[key]).up();
            }
        });
        element.up();
        return element;
    }
    static assignXmlToObject(element, obj) {
        const { children, name } = element;
        const item = { [name]: {} };
        children.forEach((node) => {
            if (typeof node === 'string') {
                item[name] = node;
            }
            else if (node.children.length > 0) {
                ChatUtils.assignXmlToObject(node, item[name]);
            }
            else {
                item[name][node.name] = ChatUtils.getText(node);
            }
        });
        return Object.assign(obj, item);
    }
    static parseExtraParams(element) {
        if (!element) {
            return null;
        }
        const extension = { attachments: [] };
        let dialogId = null;
        element.children.forEach((node) => {
            if (typeof node === 'string') {
                return;
            }
            if (node.name === 'attachment') {
                const attachment = {};
                Object.keys(node.attrs).forEach((key) => {
                    attachment[key] = key === 'size' ? parseInt(node.attrs.size) : node.attrs[key];
                });
                extension.attachments.push(attachment);
            }
            else if (node.name === 'dialog_id') {
                dialogId = ChatUtils.getElementText(element, 'dialog_id');
                extension.dialog_id = dialogId;
            }
            if (node.children.length) {
                ChatUtils.assignXmlToObject(node, extension);
            }
        });
        if (extension.attachments.length === 0) {
            delete extension.attachments;
        }
        if (extension.moduleIdentifier) {
            delete extension.moduleIdentifier;
        }
        return { extension, dialogId };
    }
    static buildErrorFromXMPPErrorStanza(errorStanza) {
        const errorElement = ChatUtils.getElement(errorStanza, 'error');
        const code = errorElement ? parseInt(ChatUtils.getAttr(errorElement, 'code')) : 0;
        const info = errorElement ? ChatUtils.getElementText(errorElement, 'text') : null;
        return { code: code, info: info };
    }
    static getUniqueId(suffix) {
        const uuid = 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, (c) => {
            const r = (Math.random() * 16) | 0;
            const v = c == 'x' ? r : (r & 0x3) | 0x8;
            return v.toString(16);
        });
        if (typeof suffix === 'string' || typeof suffix === 'number') {
            return `${uuid}:${suffix}`;
        }
        else {
            return uuid;
        }
    }
    static parseReactions(element) {
        const reactions = ChatUtils.getChildElements(element).map((node) => ({
            add: ChatUtils.getAttr(node, 'add') === 'true',
            remove: ChatUtils.getAttr(node, 'remove') === 'true',
            reaction: ChatUtils.getAttr(node, 'type'),
        }));
        return reactions.reduce((items, { add, remove, reaction }) => {
            if (add) {
                items['add'] = reaction;
            }
            else if (remove) {
                items['remove'] = reaction;
            }
            return items;
        }, {});
    }
}
exports.default = ChatUtils;
//# sourceMappingURL=ChatUtils.js.map