"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const ChatUtils_1 = __importDefault(require("./ChatUtils"));
const types_1 = require("../types");
class PrivacyList {
    xmlns = 'jabber:iq:privacy';
    helpers;
    xmppClient;
    stanzasCallbacks;
    constructor(options) {
        this.helpers = options.helpers;
        this.xmppClient = options.xmppClient;
        this.stanzasCallbacks = options.stanzasCallbacks;
    }
    async create(list) {
        return new Promise((resolve, reject) => {
            const privacyList = list.items.reduce((acc, { user_id, action, mutualBlock }) => {
                acc[user_id] = { action, mutualBlock };
                return acc;
            }, {});
            const privacyListKeys = Object.keys(privacyList);
            const iqParams = {
                type: 'set',
                from: this.helpers.userCurrentJid,
                id: ChatUtils_1.default.getUniqueId('edit'),
            };
            let iq = ChatUtils_1.default.createIqStanza(iqParams);
            iq.c('query', { xmlns: this.xmlns }).c('list', { name: list.name });
            for (let index = 0, j = 0, len = privacyListKeys.length; index < len; index++, j = j + 2) {
                const userId = privacyListKeys[index];
                const { action, mutualBlock } = privacyList[userId];
                const isMutual = (mutualBlock && action === 'deny') ?? false;
                const userJid = this.helpers.jidOrUserId(parseInt(userId, 10));
                const userMuc = this.helpers.getUserNickWithMucDomain(userId);
                this.assignPrivacyItem(iq, { order: j + 1, value: userJid, action, isMutual });
                this.assignPrivacyItem(iq, { order: j + 2, value: userMuc, action, isMutual });
            }
            this.stanzasCallbacks[iqParams.id] = (stanza) => {
                ChatUtils_1.default.isErrorStanza(stanza) ? reject(ChatUtils_1.default.buildErrorFromXMPPErrorStanza(stanza)) : resolve();
            };
            this.xmppClient.sendOnline(iq);
        });
    }
    async getList(name) {
        return new Promise((resolve, _reject) => {
            const id = ChatUtils_1.default.getUniqueId('getlist');
            const iq = ChatUtils_1.default.createIqStanza({
                type: 'get',
                from: this.helpers.userCurrentJid,
                id,
            });
            iq.c('query', { xmlns: this.xmlns }).c('list', { name });
            this.stanzasCallbacks[id] = (stanza) => {
                const query = ChatUtils_1.default.getElement(stanza, 'query');
                const list = ChatUtils_1.default.getElement(query, 'list');
                const privacyList = {
                    name: ChatUtils_1.default.getAttr(list, 'name'),
                    items: ChatUtils_1.default.getChildElements(list)
                        .map((item, index) => {
                        if (index % 2 === 0) {
                            const { action, value } = item.attrs;
                            return {
                                user_id: this.helpers.getUserIdFromJID(value),
                                action: action,
                            };
                        }
                        return null;
                    })
                        .filter((item) => item !== null),
                };
                resolve(privacyList);
                delete this.stanzasCallbacks[id];
            };
            this.xmppClient.sendOnline(iq);
        });
    }
    async update(list) {
        try {
            const currentList = await this.getList(list.name);
            const map = new Map(currentList.items.map((currentListItem) => [currentListItem.user_id, currentListItem]));
            for (const item of list.items) {
                if (item.action === types_1.PrivacyListAction.ALLOW) {
                    map.delete(item.user_id);
                }
                else {
                    const updatedItem = Object.assign({}, map.get(item.user_id), item);
                    map.set(item.user_id, updatedItem);
                }
            }
            const updatedListItems = Array.from(map.values());
            const updatedList = { name: list.name, items: updatedListItems };
            await this.create(updatedList);
        }
        catch (error) {
            throw error;
        }
    }
    async getNames() {
        return new Promise((resolve, reject) => {
            const id = ChatUtils_1.default.getUniqueId('getNames');
            const iq = ChatUtils_1.default.createIqStanza({
                type: 'get',
                from: this.helpers.userCurrentJid,
                id,
            });
            iq.c('query', { xmlns: this.xmlns });
            this.stanzasCallbacks[id] = (stanza) => {
                if (!ChatUtils_1.default.isErrorStanza(stanza)) {
                    const query = ChatUtils_1.default.getElement(stanza, 'query');
                    const defaultElement = ChatUtils_1.default.getElement(query, 'default');
                    const activeElement = ChatUtils_1.default.getElement(query, 'active');
                    const defaultName = ChatUtils_1.default.getAttr(defaultElement, 'name');
                    const activeName = ChatUtils_1.default.getAttr(activeElement, 'name');
                    const allNames = ChatUtils_1.default.getChildElements(query).map((list) => ChatUtils_1.default.getAttr(list, 'name'));
                    const namesList = {
                        default: defaultName,
                        active: activeName,
                        names: allNames,
                    };
                    resolve(namesList);
                }
                else {
                    reject(ChatUtils_1.default.buildErrorFromXMPPErrorStanza(stanza));
                }
            };
            this.xmppClient.sendOnline(iq);
        });
    }
    async delete(name) {
        return new Promise((resolve, reject) => {
            const id = ChatUtils_1.default.getUniqueId('remove');
            const iq = ChatUtils_1.default.createIqStanza({
                type: 'set',
                from: this.helpers.userCurrentJid,
                id,
            });
            iq.c('query', { xmlns: this.xmlns }).c('list', { name: name ? name : '' });
            this.stanzasCallbacks[id] = (stanza) => {
                ChatUtils_1.default.isErrorStanza(stanza) ? reject(ChatUtils_1.default.buildErrorFromXMPPErrorStanza(stanza)) : resolve();
            };
            this.xmppClient.sendOnline(iq);
        });
    }
    async setAsDefault(name) {
        return new Promise((resolve, reject) => {
            const id = ChatUtils_1.default.getUniqueId('default');
            const iq = ChatUtils_1.default.createIqStanza({
                type: 'set',
                from: this.helpers.userCurrentJid,
                id,
            });
            iq.c('query', { xmlns: this.xmlns }).c('default', name ? { name: name } : {});
            this.stanzasCallbacks[id] = (stanza) => {
                ChatUtils_1.default.isErrorStanza(stanza) ? reject(ChatUtils_1.default.buildErrorFromXMPPErrorStanza(stanza)) : resolve();
            };
            this.xmppClient.sendOnline(iq);
        });
    }
    assignPrivacyItem(iq, params) {
        const { value, action, order, isMutual } = params;
        const query = ChatUtils_1.default.getElement(iq, 'query');
        const list = ChatUtils_1.default.getElement(query, 'list');
        const item = list.c('item', { type: 'jid', value, action, order });
        if (isMutual) {
            item.up();
            if (order % 2 === 0) {
                iq.up().up();
            }
        }
        else {
            item.c('message', {}).up().c('presence-in', {}).up().c('presence-out', {}).up().c('iq', {}).up().up();
        }
    }
}
exports.default = PrivacyList;
//# sourceMappingURL=PrivacyList.js.map