"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("../types");
const config_1 = __importDefault(require("../config"));
const ChatUtils_1 = __importDefault(require("./ChatUtils"));
const Utils_1 = __importDefault(require("../Utils"));
var NonzaName;
(function (NonzaName) {
    NonzaName["ENABLE"] = "enable";
    NonzaName["ENABLED"] = "enabled";
    NonzaName["REQUEST"] = "r";
    NonzaName["ANSWER"] = "a";
})(NonzaName || (NonzaName = {}));
class StreamManagement {
    chat;
    xmlns = 'urn:xmpp:sm:3';
    enabled = false;
    clientHandledCounter = 0;
    serverHandledCounter = 0;
    messagesQueue = new Set();
    messagesQueueClearTimer;
    constructor(chat) {
        this.chat = chat;
    }
    get supported() {
        return config_1.default.chatProtocol.active === types_1.ChatProtocol.WS && config_1.default.chat.streamManagement?.enable;
    }
    start() {
        this.enabled = false;
        if (this.supported) {
            this.removeElementHandler();
            this.addElementHandler();
            this.sendNonza(NonzaName.ENABLE);
        }
    }
    stop() {
        this.enabled = false;
        if (this.supported) {
            this.removeElementHandler();
            this.markMessagesQueueAsLost();
        }
    }
    async sendAndCount(stanza, message) {
        const type = ChatUtils_1.default.getAttr(stanza, 'type');
        const body = ChatUtils_1.default.getElementText(stanza, 'body');
        const attachments = ChatUtils_1.default.getAllElements(stanza, 'attachment');
        const isMessage = stanza.name === 'message';
        const isChat = type === types_1.ChatType.CHAT || type === types_1.ChatType.GROUPCHAT;
        const isContent = Boolean(body || attachments.length);
        try {
            await this.chat.xmppClient.sendOnline(stanza);
            if (this.enabled && isMessage && isChat && isContent) {
                this.messagesQueue.add(message);
                this.sendNonza(NonzaName.REQUEST);
            }
        }
        catch {
            this.markMessageAsLost(message);
        }
    }
    elementHandler = (stanza) => {
        const [name, xmlns] = [stanza.name, ChatUtils_1.default.getAttr(stanza, 'xmlns')];
        if (xmlns === this.xmlns && name === NonzaName.ENABLED) {
            this.clientHandledCounter = 0;
            this.serverHandledCounter = 0;
            this.enabled = true;
            return;
        }
        if (xmlns !== this.xmlns)
            this.clientHandledCounter++;
        if (name === NonzaName.REQUEST) {
            this.sendNonza(NonzaName.ANSWER);
            return;
        }
        if (name === NonzaName.ANSWER) {
            const nextHandledCounter = parseInt(ChatUtils_1.default.getAttr(stanza, 'h'), 10);
            const currHandledCounter = Number.isNaN(nextHandledCounter) ? this.serverHandledCounter : nextHandledCounter;
            const diffHandledCounter = currHandledCounter - this.serverHandledCounter;
            this.serverHandledCounter = currHandledCounter;
            this.markMessagesQueueAsSent(diffHandledCounter);
            this.markMessagesQueueAsLostByTimeout();
        }
    };
    markMessageAsSent(message) {
        Utils_1.default.safeCallbackCall(this.chat.onSentMessageCallback)(null, message);
        this.messagesQueue.delete(message);
    }
    markMessageAsLost(message, clear = true) {
        Utils_1.default.safeCallbackCall(this.chat.onSentMessageCallback)(message, null);
        if (clear)
            this.messagesQueue.delete(message);
    }
    markMessagesQueueAsSent(iterations = 0) {
        const iterator = this.messagesQueue.values();
        for (let i = 0; i < iterations; i++) {
            const { done, value } = iterator.next();
            if (done)
                break;
            this.markMessageAsSent(value);
        }
    }
    markMessagesQueueAsLost() {
        clearTimeout(this.messagesQueueClearTimer);
        this.messagesQueueClearTimer = undefined;
        if (this.messagesQueue.size > 0) {
            this.messagesQueue.forEach((message) => {
                this.markMessageAsLost(message, false);
            });
            this.messagesQueue.clear();
        }
    }
    markMessagesQueueAsLostByTimeout() {
        const ackTimeout = (config_1.default.chat.streamManagement?.acknowledgementTimeout ?? 30) * 1000;
        clearTimeout(this.messagesQueueClearTimer);
        this.messagesQueueClearTimer = setTimeout(() => {
            this.markMessagesQueueAsLost();
        }, ackTimeout);
    }
    sendNonza(name) {
        const attrs = { xmlns: this.xmlns, ...(name === NonzaName.ANSWER && { h: this.clientHandledCounter }) };
        const nonza = ChatUtils_1.default.createNonza(name, attrs);
        return this.chat.xmppClient.sendOnline(nonza);
    }
    removeElementHandler() {
        this.chat.xmppClient.removeListener('element', this.elementHandler);
    }
    addElementHandler() {
        this.chat.xmppClient.on('element', this.elementHandler);
    }
}
exports.default = StreamManagement;
//# sourceMappingURL=StreamManagement.js.map