"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const platform_1 = require("../platform");
const config_1 = __importDefault(require("../config"));
const Utils_1 = __importDefault(require("../Utils"));
const WebRTCSession_1 = __importDefault(require("./WebRTCSession"));
const WebRTCSignalingProcessor_1 = __importDefault(require("./WebRTCSignalingProcessor"));
const WebRTCSignalingProvider_1 = __importDefault(require("./WebRTCSignalingProvider"));
const WebRTCHelpers_1 = __importDefault(require("./WebRTCHelpers"));
const types_1 = require("../types");
class WebRTCClient {
    connection;
    proxy;
    signalingProcessor;
    signalingProvider;
    SessionConnectionState = types_1.CallSessionConnectionState;
    PeerConnectionState = types_1.PeerConnectionState;
    CallType = types_1.CallType;
    sessions = {};
    onCallListener;
    onAcceptCallListener;
    onRejectCallListener;
    onStopCallListener;
    onInvalidEventsListener;
    onUserNotAnswerListener;
    onRemoteStreamListener;
    onSessionConnectionStateChangedListener;
    onSessionCloseListener;
    onCallStatsReport;
    onDevicesChangeListener = () => { };
    constructor(connection, proxy) {
        this.connection = connection;
        this.proxy = proxy;
        this.signalingProcessor = new WebRTCSignalingProcessor_1.default(this);
        this.signalingProvider = new WebRTCSignalingProvider_1.default(connection);
        if (platform_1.mediaDevices) {
            platform_1.mediaDevices.ondevicechange = Utils_1.default.safeCallbackCall(this.onDevicesChangeListener);
        }
    }
    get currentUserID() {
        const userJID = this.connection.jid?.toString() ?? '';
        const userID = WebRTCHelpers_1.default.getUserIdFromJID(userJID) ?? 0;
        return userID;
    }
    async getMediaDevices(kind) {
        if (platform_1.mediaDevices?.enumerateDevices) {
            const devices = await platform_1.mediaDevices.enumerateDevices();
            return kind ? devices.filter((device) => device.kind === kind) : devices;
        }
        else {
            throw new Error("No 'enumerateDevices' API supported");
        }
    }
    createNewSession(opponentsIDs, callType, options) {
        const maxBandwidth = (options?.maxBandwidth || options?.bandwidth) ?? 0;
        if (!opponentsIDs) {
            throw new Error("Can't create a session without opponentsIDs");
        }
        return this.createAndStoreSession(null, this.currentUserID, opponentsIDs, callType, maxBandwidth);
    }
    createAndStoreSession(sessionID, initiatorID, opponentsIDs, callType, maxBandwidth = 0) {
        const webRTCSession = new WebRTCSession_1.default({
            ID: sessionID,
            initiatorID,
            opponentsIDs,
            callType,
            signalingProvider: this.signalingProvider,
            currentUserID: this.currentUserID,
            maxBandwidth,
        });
        webRTCSession.onUserNotAnswerListener = this.onUserNotAnswerListener;
        webRTCSession.onRemoteStreamListener = this.onRemoteStreamListener;
        webRTCSession.onSessionConnectionStateChangedListener = this.onSessionConnectionStateChangedListener;
        webRTCSession.onSessionCloseListener = this.onSessionCloseListener;
        webRTCSession.onCallStatsReportListener = this.onCallStatsReport;
        if (webRTCSession?.ID) {
            this.sessions[webRTCSession.ID] = webRTCSession;
        }
        return webRTCSession;
    }
    clearSession(sessionId) {
        delete this.sessions[sessionId];
    }
    /// Reject call by http request
    callRejectRequest(params) {
        const ajaxParams = {
            type: types_1.ProxyMethod.POST,
            url: Utils_1.default.getUrl(config_1.default.urls.calls, 'reject'),
            data: params,
        };
        return this.proxy.ajax(ajaxParams);
    }
    /// DELEGATE (signaling)
    onCallHandler(userID, sessionID, extension) {
        const userInfo = extension.userInfo || {};
        const maxBandwidth = Number(userInfo.maxBandwidth ?? 0);
        let session = this.sessions[sessionID];
        WebRTCHelpers_1.default.trace(`onCall. userID: ${userID}, sessionID: ${sessionID}, extension: ${JSON.stringify(extension)}`);
        if (!session) {
            session = this.createAndStoreSession(sessionID, extension.callerID || 0, extension.opponentsIDs || [], extension.callType || types_1.CallType.VIDEO, maxBandwidth);
            Utils_1.default.safeCallbackCall(this.onCallListener)(session, userInfo);
        }
        session.processOnCall(userID, extension);
    }
    onAcceptHandler(userID, sessionID, extension) {
        const session = this.sessions[sessionID];
        const userInfo = extension.userInfo || {};
        WebRTCHelpers_1.default.trace(`onAccept. UserID: ${userID}, sessionID: ${sessionID}, extension: ${JSON.stringify(extension)}`);
        if (session && (session.state === types_1.CallSessionState.ACTIVE || session.state === types_1.CallSessionState.NEW)) {
            Utils_1.default.safeCallbackCall(this.onAcceptCallListener)(session, userID, userInfo);
            session.processOnAccept(userID, extension);
        }
        else {
            WebRTCHelpers_1.default.traceWarning(`Ignore 'onAccept', there is no information about session ${sessionID}`);
        }
    }
    onRejectHandler(userID, sessionID, extension) {
        const session = this.sessions[sessionID];
        WebRTCHelpers_1.default.trace(`onReject. UserID: ${userID}, sessionID: ${sessionID}, extension: ${JSON.stringify(extension)}`);
        if (session) {
            const userInfo = extension.userInfo || {};
            Utils_1.default.safeCallbackCall(this.onRejectCallListener)(session, userID, userInfo);
            session.processOnReject(userID);
        }
        else {
            WebRTCHelpers_1.default.traceWarning(`Ignore 'onReject', there is no information about session ${sessionID}`);
        }
    }
    onStopHandler(userID, sessionID, extension) {
        WebRTCHelpers_1.default.trace(`onStop. UserID: ${userID}, sessionID: ${sessionID}, extension: ${JSON.stringify(extension)}`);
        const session = this.sessions[sessionID];
        const userInfo = extension.userInfo || {};
        if (session && (session.state === types_1.CallSessionState.ACTIVE || session.state === types_1.CallSessionState.NEW)) {
            session.processOnStop(userID);
            Utils_1.default.safeCallbackCall(this.onStopCallListener)(session, userID, userInfo);
        }
        else {
            Utils_1.default.safeCallbackCall(this.onInvalidEventsListener)(session, userID, userInfo);
            WebRTCHelpers_1.default.traceWarning(`Ignore 'onStop', there is no information about session ${sessionID} by some reason`);
        }
    }
    onIceCandidatesHandler(userID, sessionID, extension) {
        const session = this.sessions[sessionID];
        WebRTCHelpers_1.default.trace(`onIceCandidates. UserID: ${userID}. SessionID: ${sessionID}. ICE candidates count: ${extension.iceCandidates?.length ?? 0}`);
        if (session) {
            if (session.state === types_1.CallSessionState.ACTIVE) {
                session.processOnIceCandidates(userID, extension);
            }
            else {
                WebRTCHelpers_1.default.traceWarning(`Ignore 'OnIceCandidates', the session ( ${sessionID} ) has invalid state`);
            }
        }
        else {
            WebRTCHelpers_1.default.traceWarning(`Ignore 'OnIceCandidates', there is no information about session ${sessionID}`);
        }
    }
    onIceRestartHandler(userID, sessionID, extension) {
        const session = this.sessions[sessionID];
        WebRTCHelpers_1.default.trace(`onIceRestart. UserID: ${userID}. SessionID: ${sessionID}. Extension: ${JSON.stringify(extension)}`);
        if (session) {
            if (session.state === types_1.CallSessionState.ACTIVE) {
                session.processOnIceRestart(userID, extension);
            }
            else {
                WebRTCHelpers_1.default.traceWarning(`Ignore 'OnIceRestart', the session (${sessionID}) has invalid state`);
            }
        }
        else {
            WebRTCHelpers_1.default.traceWarning(`Ignore 'OnIceRestart', there is no information about session ${sessionID}`);
        }
    }
    onIceRestartAcceptHandler(userID, sessionID, extension) {
        const session = this.sessions[sessionID];
        WebRTCHelpers_1.default.trace(`onIceRestartAccept. UserID: ${userID}. SessionID: ${sessionID}. Extension: ${JSON.stringify(extension)}`);
        if (session) {
            if (session.state === types_1.CallSessionState.ACTIVE) {
                session.processOnIceRestartAccept(userID, extension);
            }
            else {
                WebRTCHelpers_1.default.traceWarning(`Ignore 'onIceRestartAccept', the session (${sessionID}) has invalid state`);
            }
        }
        else {
            WebRTCHelpers_1.default.traceWarning(`Ignore 'onIceRestartAccept', there is no information about session ${sessionID}`);
        }
    }
    getListenerByName(name) {
        switch (name) {
            case types_1.CallEvent.CALL:
                return 'onCallListener';
            case types_1.CallEvent.ACCEPT:
                return 'onAcceptCallListener';
            case types_1.CallEvent.REJECT:
                return 'onRejectCallListener';
            case types_1.CallEvent.STOP:
                return 'onStopCallListener';
            case types_1.CallEvent.INVALID:
                return 'onInvalidEventsListener';
            case types_1.CallEvent.NOT_ANSWER:
                return 'onUserNotAnswerListener';
            case types_1.CallEvent.REMOTE_STREAM:
                return 'onRemoteStreamListener';
            case types_1.CallEvent.CONNECTION_STATE:
                return 'onSessionConnectionStateChangedListener';
            case types_1.CallEvent.CLOSE:
                return 'onSessionCloseListener';
            case types_1.CallEvent.STATS_REPORT:
                return 'onCallStatsReport';
            case types_1.CallEvent.DEVICES:
                return 'onDevicesChangeListener';
            default:
                return null;
        }
    }
    addListener(name, listener) {
        const listenerName = this.getListenerByName(name);
        if (listenerName) {
            this[listenerName] = listener;
        }
        return this.removeListener.bind(this, name);
    }
    removeListener(name) {
        const listenerName = this.getListenerByName(name);
        if (listenerName) {
            this[listenerName] = name === types_1.CallEvent.DEVICES ? () => { } : undefined;
        }
    }
    removeAllListeners() {
        Object.keys(this).forEach((key) => {
            if (key.startsWith('on') && key.endsWith('Listener') && typeof this[key] === 'function') {
                this[key] =
                    key === 'onDevicesChangeListener' ? () => { } : undefined;
            }
        });
    }
}
exports.default = WebRTCClient;
//# sourceMappingURL=WebRTCClient.js.map