"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const platform_1 = require("../platform");
const config_1 = __importDefault(require("../config"));
const WebRTCHelpers_1 = __importDefault(require("./WebRTCHelpers"));
const types_1 = require("../types");
class WebRTCPeerConnection {
    original;
    session;
    userID;
    type;
    remoteSDP = null;
    state = types_1.PeerConnectionState.NEW;
    iceCandidates = [];
    remoteStream = new platform_1.MediaStream();
    answerTimeInterval = 0;
    onStatusClosedChecker = null;
    dialingTimer = null;
    statsReportTimer = null;
    waitingReconnectTimeoutCallback = null;
    released = false;
    constructor(session, userID, type) {
        this.create();
        this.setup(session, userID, type);
    }
    create() {
        const rtcConfiguration = {
            iceTransportPolicy: (config_1.default.videochat.alwaysRelayCalls ? 'relay' : 'all'),
            iceServers: config_1.default.videochat.iceServers.map((item) => ({
                urls: item.urls ?? item.url,
                username: item.username,
                credential: item.credential,
            })),
        };
        this.original = platform_1.RTCPeerConnection ? new platform_1.RTCPeerConnection(rtcConfiguration) : {};
        WebRTCHelpers_1.default.trace(`new RTCPeerConnection(${JSON.stringify(rtcConfiguration, null, 2)})`);
    }
    setup(session, userID, type) {
        this.type = type;
        this.userID = userID;
        this.session = session;
        this.session.startCallTime = Date.now();
        this.original.addEventListener('track', this.onMediaTrackHandler);
        this.original.addEventListener('icecandidate', this.onIceCandidateHandler);
        this.original.addEventListener('signalingstatechange', this.onSignalingStateHandler);
        this.original.addEventListener('iceconnectionstatechange', this.onIceConnectionStateHandler);
        WebRTCHelpers_1.default.trace(`RTCPeerConnection init. ${this.toString()}`);
    }
    onMediaTrackHandler = (event) => {
        this.remoteStream.addTrack(event.track);
        if ((this.session.callType == types_1.CallType.VIDEO && this.remoteStream.getVideoTracks().length) ||
            (this.session.callType == types_1.CallType.AUDIO && this.remoteStream.getAudioTracks().length)) {
            this.session.onRemoteStream(this.userID, this.remoteStream);
        }
        this.getWrappedStats();
    };
    onIceCandidateHandler = (event) => {
        if (event.candidate) {
            const { sdpMLineIndex, sdpMid, candidate } = event.candidate;
            const iceCandidate = { sdpMLineIndex, sdpMid, candidate };
            if (this.signalingState === 'stable' && this.original.canTrickleIceCandidates) {
                this.session.processIceCandidates(this.userID, [iceCandidate]);
            }
            else {
                this.iceCandidates.push(iceCandidate);
            }
        }
    };
    onSignalingStateHandler = async () => {
        WebRTCHelpers_1.default.trace(`onSignalingStateHandler: ${this.signalingState}`);
        if (this.signalingState === 'stable' && this.iceCandidates.length > 0) {
            this.session.processIceCandidates(this.userID, this.iceCandidates);
            this.iceCandidates.length = 0;
        }
    };
    /** handler of remote media stream */
    onIceConnectionStateHandler = () => {
        WebRTCHelpers_1.default.trace(`onIceConnectionStateHandler: ${this.iceConnectionState}`);
        if (this.onStatusClosedChecker && WebRTCHelpers_1.default.getVersionSafari >= 11) {
            clearTimeout(this.onStatusClosedChecker);
        }
        switch (this.iceConnectionState) {
            case 'checking':
                this.state = types_1.PeerConnectionState.CHECKING;
                this.session.onSessionConnectionStateChanged(this.userID, types_1.CallSessionConnectionState.CONNECTING);
                break;
            case 'connected':
                this.state = types_1.PeerConnectionState.CONNECTED;
                this.clearWaitingReconnectTimer();
                this.session.onSessionConnectionStateChanged(this.userID, types_1.CallSessionConnectionState.CONNECTED);
                break;
            case 'completed':
                this.state = types_1.PeerConnectionState.COMPLETED;
                this.clearWaitingReconnectTimer();
                this.session.onSessionConnectionStateChanged(this.userID, types_1.CallSessionConnectionState.COMPLETED);
                break;
            case 'failed':
                this.state = types_1.PeerConnectionState.FAILED;
                this.session.onSessionConnectionStateChanged(this.userID, types_1.CallSessionConnectionState.FAILED);
                break;
            case 'disconnected':
                this.state = types_1.PeerConnectionState.DISCONNECTED;
                this.startWaitingReconnectTimer();
                this.session.onSessionConnectionStateChanged(this.userID, types_1.CallSessionConnectionState.DISCONNECTED);
                if (WebRTCHelpers_1.default.getVersionSafari >= 11) {
                    this.onStatusClosedChecker = setTimeout(() => {
                        this.onIceConnectionStateHandler(); // repeat to get status "closed"
                    }, 500);
                }
                break;
            case 'closed': // TODO: this state doesn't fires on Safari 11
                this.state = types_1.PeerConnectionState.CLOSED;
                this.clearWaitingReconnectTimer();
                this.session.onSessionConnectionStateChanged(this.userID, types_1.CallSessionConnectionState.CLOSED);
                break;
            default:
                break;
        }
    };
    clearWaitingReconnectTimer() {
        if (this.waitingReconnectTimeoutCallback) {
            WebRTCHelpers_1.default.trace('clearWaitingReconnectTimer');
            clearTimeout(this.waitingReconnectTimeoutCallback);
            this.waitingReconnectTimeoutCallback = null;
        }
    }
    startWaitingReconnectTimer() {
        const timeout = config_1.default.videochat.disconnectTimeInterval * 1000;
        WebRTCHelpers_1.default.trace(`startWaitingReconnectTimer, timeout: ${timeout}`);
        this.waitingReconnectTimeoutCallback = setTimeout(() => {
            WebRTCHelpers_1.default.trace('waitingReconnectTimeoutCallback');
            if (this.waitingReconnectTimeoutCallback) {
                clearTimeout(this.waitingReconnectTimeoutCallback);
            }
            this.release();
            this.session.closeSessionIfAllConnectionsClosed();
        }, timeout);
    }
    clearStatsReportTimer() {
        if (this.statsReportTimer) {
            clearInterval(this.statsReportTimer);
            this.statsReportTimer = null;
        }
    }
    async addCandidates(iceCandidates) {
        for (const iceCandidate of iceCandidates) {
            if (iceCandidate?.candidate) {
                try {
                    await this.addIceCandidate(iceCandidate);
                    WebRTCHelpers_1.default.trace(`'addIceCandidate' success: ${this.signalingState}`);
                }
                catch (error) {
                    WebRTCHelpers_1.default.traceError(`'addIceCandidate' error: ${error}`);
                }
            }
        }
    }
    release() {
        this.clearDialingTimer();
        this.clearStatsReportTimer();
        this.close();
        if (WebRTCHelpers_1.default.getVersionSafari >= 11) {
            this.onIceConnectionStateHandler(); // TODO: 'closed' state doesn't fires on Safari 11 (do it manually)
        }
        this.original.removeEventListener('track', this.onMediaTrackHandler);
        this.original.removeEventListener('icecandidate', this.onIceCandidateHandler);
        this.original.removeEventListener('signalingstatechange', this.onSignalingStateHandler);
        this.original.removeEventListener('iceconnectionstatechange', this.onIceConnectionStateHandler);
        this.released = true;
        WebRTCHelpers_1.default.trace(`RTCPeerConnection closed. ${this.toString()}`);
    }
    clearDialingTimer() {
        if (this.dialingTimer) {
            WebRTCHelpers_1.default.trace('clearDialingTimer');
            clearInterval(this.dialingTimer);
            this.dialingTimer = null;
            this.answerTimeInterval = 0;
        }
    }
    async getAndSetLocalSessionDescription(maxBandwidth, iceRestart = false) {
        this.state = types_1.PeerConnectionState.CONNECTING;
        try {
            const offerOptions = iceRestart ? { iceRestart } : undefined;
            const sessionDescription = this.type === 'offer' ? await this.createOffer(offerOptions) : await this.createAnswer();
            await this.setLocalDescription(sessionDescription);
            this.setRTCRtpSenderMaxBandwidth(maxBandwidth);
            return sessionDescription;
        }
        catch (error) {
            throw error;
        }
    }
    setRTCRtpSenderMaxBandwidth(maxBandwidth) {
        const senders = this.getSenders() || [];
        const maxBitrate = maxBandwidth ? Math.round(maxBandwidth * 1000) : 0;
        senders.forEach((sender) => {
            if (sender.track?.kind === types_1.MediaType.VIDEO) {
                const params = sender.getParameters();
                params.encodings ??= [{}];
                params.encodings.forEach((encoding) => {
                    if (maxBitrate > 0) {
                        encoding.maxBitrate = maxBitrate;
                    }
                    else {
                        delete encoding.maxBitrate;
                    }
                });
                sender
                    .setParameters(params)
                    .then(() => {
                    WebRTCHelpers_1.default.trace(`Set 'maxBandwidth' for ${this.userID}: ${maxBandwidth > 0 ? `${maxBandwidth} kbps` : 'unlimited'}`);
                })
                    .catch((error) => {
                    WebRTCHelpers_1.default.traceWarning(`Set 'maxBandwidth' for ${this.userID}: ${error}`);
                });
            }
        });
    }
    startDialingTimer(extension, withOnNotAnswerCallback) {
        const dialingTimeInterval = config_1.default.videochat.dialingTimeInterval * 1000;
        WebRTCHelpers_1.default.trace(`startDialingTimer, dialingTimeInterval: ${JSON.stringify(dialingTimeInterval)}`);
        const dialingCallback = (extension, withOnNotAnswerCallback, skipIncrement) => {
            if (!skipIncrement) {
                this.answerTimeInterval += config_1.default.videochat.dialingTimeInterval * 1000;
            }
            WebRTCHelpers_1.default.trace(`dialingCallback, extension: ${JSON.stringify(extension)}`);
            if (this.answerTimeInterval >= config_1.default.videochat.answerTimeInterval * 1000) {
                this.clearDialingTimer();
                if (withOnNotAnswerCallback) {
                    this.session.processOnNotAnswer(this);
                }
            }
            else {
                this.session.processCall(this, extension);
            }
        };
        this.dialingTimer = setInterval(dialingCallback, dialingTimeInterval, extension, withOnNotAnswerCallback, false);
        dialingCallback(extension, withOnNotAnswerCallback, true); // call for the 1st time
    }
    async setRemoteSessionDescription(type, sdp) {
        const desc = new platform_1.RTCSessionDescription({ sdp, type });
        return this.setRemoteDescription(desc);
    }
    getWrappedStats() {
        if (config_1.default.videochat.statsReportTimeInterval) {
            let lastResult;
            WebRTCHelpers_1.default.trace('Stats tracker has been started');
            this.statsReportTimer = setInterval(() => {
                this.getStatsCustom(lastResult, (results, lastResults) => {
                    lastResult = lastResults;
                    this.session.onCallStatsReport(this.userID, results);
                }, (err) => {
                    WebRTCHelpers_1.default.traceError(`Get stats error. ${err.name}: ${err.message}`);
                    this.session.onCallStatsReport(this.userID, null, err);
                });
            }, config_1.default.videochat.statsReportTimeInterval * 1000);
        }
    }
    set sdpRemote(sdp) {
        if (sdp) {
            this.remoteSDP = sdp;
        }
    }
    get sdpRemote() {
        return this.remoteSDP;
    }
    toString() {
        return `sessionID: ${this.session.ID}, userID: ${this.userID}, type: ${this.type}, state: ${this.state}`;
    }
    getStatsCustom(lastResults, onSuccess, onFail) {
        const obj = { audio: {}, video: {}, candidate: {} };
        const statistic = { local: Object.assign({}, obj), remote: Object.assign({}, obj) };
        if (WebRTCHelpers_1.default.getVersionFirefox) {
            const localStream = this.session.localStream;
            const localVideoSettings = localStream?.getVideoTracks()[0].getSettings();
            if (localVideoSettings) {
                statistic.local.video.frameHeight = localVideoSettings.height;
                statistic.local.video.frameWidth = localVideoSettings.width;
            }
        }
        this.getStats().then((results) => {
            results.forEach((result) => {
                let item;
                if (result.bytesReceived && result.type === 'inbound-rtp') {
                    item = statistic.remote[result.mediaType];
                    item.bitrate = this.getBitratePerSecond(result, lastResults, false);
                    item.bytesReceived = result.bytesReceived;
                    item.packetsReceived = result.packetsReceived;
                    item.timestamp = result.timestamp;
                    if (result.mediaType === types_1.MediaType.VIDEO && result.framerateMean) {
                        item.framesPerSecond = Math.round(result.framerateMean * 10) / 10;
                    }
                }
                else if (result.bytesSent && result.type === 'outbound-rtp') {
                    item = statistic.local[result.mediaType];
                    item.bitrate = this.getBitratePerSecond(result, lastResults, true);
                    item.bytesSent = result.bytesSent;
                    item.packetsSent = result.packetsSent;
                    item.timestamp = result.timestamp;
                    if (result.mediaType === types_1.MediaType.VIDEO && result.framerateMean) {
                        item.framesPerSecond = Math.round(result.framerateMean * 10) / 10;
                    }
                }
                else if (result.type === 'local-candidate') {
                    item = statistic.local.candidate;
                    if (result.candidateType === 'host' && result.mozLocalTransport === 'udp' && result.transport === 'udp') {
                        item.protocol = result.transport;
                        item.ip = result.ipAddress;
                        item.port = result.portNumber;
                    }
                    else if (!WebRTCHelpers_1.default.getVersionFirefox) {
                        item.protocol = result.protocol;
                        item.ip = result.ip;
                        item.port = result.port;
                    }
                }
                else if (result.type === 'remote-candidate') {
                    item = statistic.remote.candidate;
                    item.protocol = result.protocol || result.transport;
                    item.ip = result.ip || result.ipAddress;
                    item.port = result.port || result.portNumber;
                }
                else if (result.type === 'track' && result.kind === types_1.MediaType.VIDEO && !WebRTCHelpers_1.default.getVersionFirefox) {
                    if (result.remoteSource) {
                        item = statistic.remote.video;
                        item.frameHeight = result.frameHeight;
                        item.frameWidth = result.frameWidth;
                        item.framesPerSecond = this.getFramesPerSecond(result, lastResults, false);
                    }
                    else {
                        item = statistic.local.video;
                        item.frameHeight = result.frameHeight;
                        item.frameWidth = result.frameWidth;
                        item.framesPerSecond = this.getFramesPerSecond(result, lastResults, true);
                    }
                }
            });
            onSuccess(statistic, results);
        }, onFail);
    }
    getBitratePerSecond(result, lastResults, isLocal) {
        const lastResult = lastResults.get(result.id);
        const seconds = lastResult ? (result.timestamp - lastResult.timestamp) / 1000 : 5;
        const kilo = 1024;
        const bit = 8;
        const bitrate = !lastResult
            ? 0
            : isLocal
                ? (bit * (result.bytesSent - lastResult.bytesSent)) / (kilo * seconds)
                : (bit * (result.bytesReceived - lastResult.bytesReceived)) / (kilo * seconds);
        return Math.round(bitrate);
    }
    getFramesPerSecond(result, lastResults, isLocal) {
        const lastResult = lastResults && lastResults.get(result.id);
        const seconds = lastResult ? (result.timestamp - lastResult.timestamp) / 1000 : 5;
        const framesPerSecond = !lastResult
            ? 0
            : isLocal
                ? (result.framesSent - lastResult.framesSent) / seconds
                : (result.framesReceived - lastResult.framesReceived) / seconds;
        return Math.round(framesPerSecond * 10) / 10;
    }
    // original RTCPeerConnection methods
    close() {
        this.original.close?.();
    }
    addTrack(track, stream) {
        return this.original.addTrack?.(track, stream) ?? null;
    }
    getSenders() {
        return this.original.getSenders?.() ?? null;
    }
    async createOffer(options) {
        return (await this.original.createOffer?.(options)) ?? null;
    }
    async createAnswer(options) {
        return (await this.original.createAnswer?.(options)) ?? null;
    }
    async getStats(selector) {
        return (await this.original.getStats?.(selector)) ?? null;
    }
    async addIceCandidate(candidate) {
        await this.original.addIceCandidate?.(candidate);
    }
    async setRemoteDescription(description) {
        await this.original.setRemoteDescription?.(description);
    }
    async setLocalDescription(description) {
        await this.original.setLocalDescription?.(description);
    }
    get localDescription() {
        return this.original.localDescription ?? null;
    }
    get signalingState() {
        return this.original.signalingState ?? 'closed';
    }
    get iceConnectionState() {
        return this.original.iceConnectionState ?? 'closed';
    }
}
exports.default = WebRTCPeerConnection;
//# sourceMappingURL=WebRTCPeerConnection.js.map