"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const platform_1 = require("../platform");
const config_1 = __importDefault(require("../config"));
const WebRTCPeerConnection_1 = __importDefault(require("./WebRTCPeerConnection"));
const WebRTCHelpers_1 = __importDefault(require("./WebRTCHelpers"));
const Utils_1 = __importDefault(require("../Utils"));
const types_1 = require("../types");
class WebRTCSession {
    ID = null;
    state = types_1.CallSessionState.NEW;
    callType;
    initiatorID;
    currentUserID;
    opponentsIDs = [];
    maxBandwidth = 0;
    peerConnections = {};
    localStream;
    mediaParams = {};
    signalingProvider;
    answerTimer = null;
    waitingOfferOrAnswerTimer = null;
    startCallTime = 0;
    acceptCallTime = 0;
    onUserNotAnswerListener;
    onRemoteStreamListener;
    onSessionCloseListener;
    onCallStatsReportListener;
    onSessionConnectionStateChangedListener;
    constructor(params) {
        Object.assign(this, params, { ID: params.ID ?? generateUUID() });
    }
    async getDisplayMedia(params) {
        if (!platform_1.mediaDevices.getDisplayMedia) {
            throw new Error("Your environment does not support 'getDisplayMedia' API");
        }
        const elementId = params && params.elementId;
        const options = params && params.options;
        const mediaParams = { ...params };
        this.mediaParams = mediaParams;
        delete mediaParams.elementId;
        delete mediaParams.options;
        try {
            const stream = await platform_1.mediaDevices.getDisplayMedia(mediaParams);
            const localStream = this.upsertStream(stream, elementId, options);
            return localStream;
        }
        catch (error) {
            throw error;
        }
    }
    async getUserMedia(params) {
        const elementId = params && params.elementId;
        const options = params && params.options;
        const mediaParams = { ...params };
        this.mediaParams = mediaParams;
        delete mediaParams.elementId;
        delete mediaParams.options;
        try {
            const stream = await platform_1.mediaDevices.getUserMedia(mediaParams);
            const localStream = this.upsertStream(stream, elementId, options);
            return localStream;
        }
        catch (error) {
            throw error;
        }
    }
    upsertStream(stream, elementId, options) {
        const shouldUpdateCurrentStream = !!this.localStream;
        if (shouldUpdateCurrentStream) {
            this.replaceTracks(stream);
        }
        else {
            this.localStream = stream;
        }
        if (!this.localStream) {
            throw new Error('Local stream is not defined');
        }
        if (elementId) {
            if (shouldUpdateCurrentStream) {
                this.detachMediaStream(elementId, options);
            }
            this.attachMediaStream(elementId, this.localStream, options);
        }
        return this.localStream;
    }
    replaceTracks(stream) {
        this.localStream?.getTracks().forEach((localTrack) => {
            if (localTrack.kind === types_1.MediaType.AUDIO && stream.getAudioTracks().length === 0) {
                return;
            }
            else {
                localTrack.stop();
                this.localStream?.removeTrack(localTrack);
            }
        });
        stream.getTracks().forEach((newTrack) => {
            if (newTrack.kind === types_1.MediaType.AUDIO) {
                newTrack.enabled = this.localStream?.getAudioTracks().every(({ enabled }) => enabled) ?? true;
            }
            else {
                newTrack.enabled = this.localStream?.getVideoTracks().every(({ enabled }) => enabled) ?? true;
            }
            if (newTrack) {
                Object.values(this.peerConnections).forEach((pc) => {
                    pc.getSenders()
                        .find(({ track }) => newTrack.kind === track?.kind)
                        ?.replaceTrack(newTrack);
                });
                this.localStream?.addTrack(newTrack);
            }
        });
        if (!this.localStream) {
            throw new Error('Local stream is not defined');
        }
        return this.localStream;
    }
    async setMaxBandwidth(maxBandwidth) {
        const peers = this.peerConnections;
        const peersKeys = Object.keys(peers);
        if (peersKeys.length < 1) {
            WebRTCHelpers_1.default.trace("No 'RTCPeerConnection' to set 'maxBandwidth'");
            return;
        }
        return Promise.all(peersKeys.map((userID) => peers[Number(userID)].setRTCRtpSenderMaxBandwidth(maxBandwidth)));
    }
    connectionStateForUser(userID) {
        const peerConnection = this.peerConnections[userID];
        return peerConnection ? peerConnection.state : null;
    }
    attachMediaStream(elementId, stream, opt) {
        const mediaElement = document.getElementById(elementId);
        if ('srcObject' in mediaElement) {
            mediaElement.srcObject = stream;
            mediaElement.style.transform = opt?.mirror ? 'scaleX(-1)' : 'none';
            if (typeof opt?.muted === 'boolean') {
                mediaElement.muted = opt?.muted;
            }
            mediaElement.onloadedmetadata = (_event) => {
                mediaElement.play();
            };
        }
        else {
            throw new Error(`Unable to attach media stream, element #${elementId} is undefined`);
        }
    }
    detachMediaStream(elementId, opt) {
        const mediaElement = document.getElementById(elementId);
        if ('srcObject' in mediaElement) {
            mediaElement.pause();
            mediaElement.srcObject = null;
            mediaElement.style.transform = opt?.mirror ? 'scaleX(-1)' : 'none';
            if (typeof opt?.muted === 'boolean') {
                mediaElement.muted = opt?.muted;
            }
        }
        else {
            throw new Error(`Unable to attach media stream, element #${elementId} is undefined`);
        }
    }
    async switchMediaTracks(options = {}) {
        Utils_1.default.DLog('switchMediaTracks:', options);
        [types_1.MediaType.AUDIO, types_1.MediaType.VIDEO].forEach((type) => {
            const device = type === types_1.MediaType.AUDIO ? options.audio : type === types_1.MediaType.VIDEO ? options.video : {};
            const deviceId = typeof device === 'string' ? device : device?.deviceId;
            if (deviceId) {
                if (typeof this.mediaParams[type] !== 'object') {
                    this.mediaParams[type].deviceId = deviceId;
                }
                else {
                    this.mediaParams[type] = { deviceId };
                }
            }
        });
        try {
            const constrains = { audio: this.mediaParams.audio || false, video: this.mediaParams.video || false };
            const stream = await platform_1.mediaDevices.getUserMedia(constrains);
            const localStream = this.replaceTracks(stream);
            return localStream;
        }
        catch (error) {
            throw error;
        }
    }
    call(extension = {}) {
        const ext = prepareExtension(extension);
        WebRTCHelpers_1.default.trace(`Call, extension: ${JSON.stringify(ext)}`);
        this.state = types_1.CallSessionState.ACTIVE;
        this.maxBandwidth = Number(ext.userInfo?.maxBandwidth ?? 0);
        this.opponentsIDs.forEach((userID) => {
            this.callInternal(userID, ext, true);
        });
    }
    async callInternal(userID, extension, withOnNotAnswerCallback) {
        const pc = new WebRTCPeerConnection_1.default(this, userID, 'offer');
        this.localStream?.getTracks().forEach((track) => {
            pc.addTrack(track, this.localStream);
        });
        this.peerConnections[userID] = pc;
        try {
            await pc.getAndSetLocalSessionDescription(this.maxBandwidth);
            WebRTCHelpers_1.default.trace('getAndSetLocalSessionDescription success');
            pc.startDialingTimer(extension, withOnNotAnswerCallback); // call requests to user
        }
        catch (error) {
            WebRTCHelpers_1.default.traceError(`getAndSetLocalSessionDescription error: ${error}`);
        }
    }
    accept(extension = {}) {
        const ext = prepareExtension(extension);
        WebRTCHelpers_1.default.trace(`Accept, extension: ${JSON.stringify(ext)}`);
        if (this.state === types_1.CallSessionState.ACTIVE) {
            WebRTCHelpers_1.default.traceError("Can't accept, the session is already active, return");
            return;
        }
        if (this.state === types_1.CallSessionState.CLOSED) {
            WebRTCHelpers_1.default.traceError("Can't accept, the session is already closed, return");
            this.stop({});
            return;
        }
        this.state = types_1.CallSessionState.ACTIVE;
        this.acceptCallTime = Date.now();
        this.clearAnswerTimer();
        this.acceptInternal(this.initiatorID, ext);
        // group call
        const opponentsIDs = this.opponentsIDs.filter((userID) => userID !== this.currentUserID);
        if (opponentsIDs.length > 0) {
            this.startWaitingOfferOrAnswerTimer();
            opponentsIDs.forEach((opponentID) => {
                if (this.currentUserID > opponentID) {
                    this.callInternal(opponentID, ext, true);
                }
            });
        }
    }
    async acceptInternal(userID, extension) {
        const pc = this.peerConnections[userID];
        if (!pc || !pc.sdpRemote) {
            WebRTCHelpers_1.default.traceError("Can't accept the call, there is no information about peer connection by some reason");
            return;
        }
        this.localStream?.getTracks().forEach((track) => {
            pc.addTrack(track, this.localStream);
        });
        try {
            await pc.setRemoteSessionDescription('offer', pc.sdpRemote);
            WebRTCHelpers_1.default.trace('setRemoteSessionDescription - success');
            await pc.getAndSetLocalSessionDescription(this.maxBandwidth);
            WebRTCHelpers_1.default.trace('getAndSetLocalSessionDescription - success');
            const ext = Object.assign({}, extension, {
                sessionID: this.ID,
                callType: this.callType,
                callerID: this.initiatorID,
                opponentsIDs: this.opponentsIDs,
                sdp: pc.localDescription.sdp,
            });
            this.signalingProvider.sendMessage(userID, ext, types_1.CallSignalingType.ACCEPT);
        }
        catch (error) {
            WebRTCHelpers_1.default.traceError(`[acceptInternal] Error: ${error}`);
        }
    }
    reject(extension = {}) {
        const ext = prepareExtension(extension);
        WebRTCHelpers_1.default.trace(`Reject, extension: ${JSON.stringify(ext.userInfo)}`);
        this.state = types_1.CallSessionState.REJECTED;
        this.clearAnswerTimer();
        Object.assign(ext, {
            sessionID: this.ID,
            callType: this.callType,
            callerID: this.initiatorID,
            opponentsIDs: this.opponentsIDs,
        });
        Object.keys(this.peerConnections).forEach((userID) => {
            this.signalingProvider.sendMessage(Number(userID), ext, types_1.CallSignalingType.REJECT);
        });
        this.close();
    }
    stop(extension = {}) {
        const ext = prepareExtension(extension);
        WebRTCHelpers_1.default.trace(`Stop, extension: ${JSON.stringify(ext.userInfo)}`);
        this.state = types_1.CallSessionState.STOPPED;
        if (this.answerTimer) {
            this.clearAnswerTimer();
        }
        Object.assign(ext, {
            sessionID: this.ID,
            callType: this.callType,
            callerID: this.initiatorID,
            opponentsIDs: this.opponentsIDs,
        });
        Object.keys(this.peerConnections).forEach((userID) => {
            this.signalingProvider.sendMessage(Number(userID), ext, types_1.CallSignalingType.STOP);
        });
        this.close();
    }
    canInitiateIceRestart(userID) {
        return this.peerConnections[userID]?.type === 'offer';
    }
    async iceRestart(userID) {
        const pc = this.peerConnections[userID];
        if (!pc) {
            WebRTCHelpers_1.default.traceError("Can't restart ice, there is no information about peer connection by some reason");
            return;
        }
        try {
            const { sdp } = await pc.getAndSetLocalSessionDescription(this.maxBandwidth, true);
            const ext = { sessionID: this.ID ?? '', sdp };
            this.signalingProvider.sendMessage(userID, ext, types_1.CallSignalingType.RESTART);
            WebRTCHelpers_1.default.trace('[iceRestart] OK');
        }
        catch (error) {
            WebRTCHelpers_1.default.traceError(`[iceRestart] Error: ${error}`);
        }
    }
    processOnCall(callerID, extension) {
        const opponentsIDs = [this.initiatorID, ...this.opponentsIDs.filter((userID) => userID !== this.currentUserID)];
        opponentsIDs.forEach((opponentID) => {
            const pc = this.peerConnections[opponentID];
            if (pc) {
                if (opponentID === callerID) {
                    pc.sdpRemote = extension.sdp;
                    /** The group call logic starts here */
                    if (callerID !== this.initiatorID && this.state === types_1.CallSessionState.ACTIVE) {
                        this.acceptInternal(callerID, extension);
                    }
                }
            }
            else {
                /** create peer connections for each opponent */
                const rtcSdpType = opponentID !== callerID && this.currentUserID > opponentID ? 'offer' : 'answer';
                const pc = new WebRTCPeerConnection_1.default(this, opponentID, rtcSdpType);
                this.peerConnections[opponentID] = pc;
                if (opponentID === callerID) {
                    pc.sdpRemote = extension.sdp;
                    this.startAnswerTimer();
                }
            }
        });
    }
    async processOnAccept(userID, extension) {
        const pc = this.peerConnections[userID];
        if (!pc) {
            WebRTCHelpers_1.default.traceWarning("Ignore 'OnAccept', there is no information about peer connection by some reason");
        }
        try {
            pc.clearDialingTimer();
            await pc.setRemoteSessionDescription('answer', extension.sdp);
            WebRTCHelpers_1.default.trace("'setRemoteSessionDescription' success");
        }
        catch (error) {
            WebRTCHelpers_1.default.traceError(`'setRemoteSessionDescription' error: ${error}`);
        }
    }
    processOnReject(userID) {
        const pc = this.peerConnections[userID];
        this.clearWaitingOfferOrAnswerTimer();
        if (pc) {
            pc.release();
        }
        else {
            WebRTCHelpers_1.default.traceWarning("Ignore 'OnReject', there is no information about peer connection by some reason");
        }
        this.closeSessionIfAllConnectionsClosed();
    }
    processOnStop(userID) {
        this.clearAnswerTimer();
        /** drop the call if the initiator did it */
        if (userID === this.initiatorID) {
            const pcKeys = Object.keys(this.peerConnections);
            if (pcKeys.length > 0) {
                pcKeys.forEach((key) => {
                    this.peerConnections[Number(key)].release();
                });
            }
            else {
                WebRTCHelpers_1.default.traceWarning("Ignore 'OnStop', there is no information about peer connections by some reason");
            }
        }
        else {
            const pc = this.peerConnections[userID];
            if (pc) {
                pc.release();
            }
            else {
                WebRTCHelpers_1.default.traceWarning('Ignore "OnStop", there is no information about peer connection by some reason');
            }
        }
        this.closeSessionIfAllConnectionsClosed();
    }
    async processOnIceCandidates(userID, extension) {
        const pc = this.peerConnections[userID];
        if (!pc) {
            WebRTCHelpers_1.default.traceWarning('Ignore "OnIceCandidates", there is no information about peer connection by some reason');
        }
        if (extension.iceCandidates) {
            await pc.addCandidates(extension.iceCandidates);
        }
    }
    async processOnIceRestart(userID, extension) {
        const pc = this.peerConnections[userID];
        if (!pc) {
            WebRTCHelpers_1.default.traceWarning('Ignore "OnIceRestart", there is no information about peer connection by some reason');
        }
        try {
            await pc.setRemoteSessionDescription('offer', extension.sdp);
            const { sdp } = await pc.getAndSetLocalSessionDescription(this.maxBandwidth);
            const ext = { sessionID: this.ID ?? '', sdp };
            this.signalingProvider.sendMessage(userID, ext, types_1.CallSignalingType.RESTART_ACCEPT);
            WebRTCHelpers_1.default.trace('[processOnIceRestart] Success');
        }
        catch (error) {
            WebRTCHelpers_1.default.traceError(`[processOnIceRestart] Error: ${error}`);
        }
    }
    async processOnIceRestartAccept(userID, extension) {
        const pc = this.peerConnections[userID];
        if (!pc) {
            WebRTCHelpers_1.default.traceWarning('Ignore "OnIceRestartAccept", there is no information about peer connection by some reason');
        }
        try {
            await pc.setRemoteSessionDescription('answer', extension.sdp);
            WebRTCHelpers_1.default.trace('[processOnIceRestartAccept] Success');
        }
        catch (error) {
            WebRTCHelpers_1.default.traceError(`[processOnIceRestartAccept] Error: ${error}`);
        }
    }
    processCall(peerConnection, extension = {}) {
        const ext = Object.assign({ userInfo: {} }, extension, {
            sessionID: this.ID,
            callType: this.callType,
            callerID: this.initiatorID,
            opponentsIDs: this.opponentsIDs,
            sdp: peerConnection.localDescription.sdp,
        });
        this.signalingProvider.sendMessage(peerConnection.userID, ext, types_1.CallSignalingType.CALL);
    }
    processIceCandidates(userID, iceCandidates) {
        const extension = {
            sessionID: this.ID ?? '',
            callType: this.callType,
            callerID: this.initiatorID,
            opponentsIDs: this.opponentsIDs,
        };
        this.signalingProvider.sendCandidate(userID, iceCandidates, extension);
    }
    processOnNotAnswer(peerConnection) {
        WebRTCHelpers_1.default.trace(`Answer timeout callback for session ${this.ID} for user ${peerConnection.userID}`);
        this.clearWaitingOfferOrAnswerTimer();
        peerConnection.release();
        Utils_1.default.safeCallbackCall(this.onUserNotAnswerListener)(this, peerConnection.userID);
        this.closeSessionIfAllConnectionsClosed();
    }
    onRemoteStream(userID, stream) {
        Utils_1.default.safeCallbackCall(this.onRemoteStreamListener)(this, userID, stream);
    }
    onCallStatsReport(userID, stats, error) {
        Utils_1.default.safeCallbackCall(this.onCallStatsReportListener)(this, userID, stats, error);
    }
    onSessionConnectionStateChanged(userID, connectionState) {
        Utils_1.default.safeCallbackCall(this.onSessionConnectionStateChangedListener)(this, userID, connectionState);
    }
    close() {
        Object.values(this.peerConnections).forEach((pc) => {
            try {
                pc.release();
            }
            catch (e) {
                Utils_1.default.DLog(`Peer close error: ${e}`);
            }
        });
        this.closeLocalMediaStream();
        this.state = types_1.CallSessionState.CLOSED;
        Utils_1.default.safeCallbackCall(this.onSessionCloseListener)(this);
    }
    closeSessionIfAllConnectionsClosed() {
        let isAllClosed = true;
        Object.values(this.peerConnections).forEach((pc) => {
            let isClosed = false;
            try {
                /*
                 * TODO:
                 * Uses RTCPeerConnection.signalingState instead RTCPeerConnection.iceConnectionState,
                 * because state 'closed' comes after few time from Safari, but signaling state comes instantly
                 */
                isClosed = pc.iceConnectionState === 'closed' || pc.signalingState === 'closed' || pc.released;
            }
            catch (err) {
                WebRTCHelpers_1.default.traceError(err);
                // need to set peerState to 'closed' on error. FF will crashed without this part.
                isClosed = true;
            }
            isAllClosed = isClosed;
        });
        WebRTCHelpers_1.default.trace(`All peer connections closed: ${isAllClosed}`);
        if (isAllClosed) {
            this.closeLocalMediaStream();
            Utils_1.default.safeCallbackCall(this.onSessionCloseListener)(this);
            this.state = types_1.CallSessionState.CLOSED;
        }
    }
    closeLocalMediaStream() {
        if (this.localStream) {
            this.localStream.getTracks().forEach((track) => track.stop());
            this.localStream = undefined;
            this.mediaParams = {};
        }
    }
    mute(type) {
        this.muteStream(false, type);
    }
    unmute(type) {
        this.muteStream(true, type);
    }
    muteStream(enabled, type = types_1.MediaType.NONE) {
        const tracks = type === types_1.MediaType.AUDIO
            ? this.localStream?.getAudioTracks()
            : type === types_1.MediaType.VIDEO
                ? this.localStream?.getVideoTracks()
                : [];
        tracks?.forEach((track) => {
            track.enabled = enabled;
        });
    }
    clearAnswerTimer() {
        if (this.answerTimer) {
            WebRTCHelpers_1.default.trace('clearAnswerTimer');
            clearTimeout(this.answerTimer);
            this.answerTimer = null;
        }
    }
    startAnswerTimer() {
        WebRTCHelpers_1.default.trace('startAnswerTimer');
        this.answerTimer = setTimeout(() => {
            WebRTCHelpers_1.default.trace('answerTimeoutCallback');
            this.close();
            this.answerTimer = null;
        }, config_1.default.videochat.answerTimeInterval * 1000);
    }
    clearWaitingOfferOrAnswerTimer() {
        if (this.waitingOfferOrAnswerTimer) {
            WebRTCHelpers_1.default.trace('clearWaitingOfferOrAnswerTimer');
            clearTimeout(this.waitingOfferOrAnswerTimer);
            this.waitingOfferOrAnswerTimer = null;
        }
    }
    startWaitingOfferOrAnswerTimer() {
        const timeWait = (this.acceptCallTime - this.startCallTime) / 1000;
        const timeDiff = Math.min(timeWait, config_1.default.videochat.answerTimeInterval / 2);
        const timeout = Math.max(config_1.default.videochat.answerTimeInterval - timeDiff, 1);
        WebRTCHelpers_1.default.trace(`startWaitingOfferOrAnswerTimer, timeout: ${timeout}`);
        this.waitingOfferOrAnswerTimer = setTimeout(() => {
            WebRTCHelpers_1.default.trace('waitingOfferOrAnswerTimeoutCallback');
            Object.values(this.peerConnections).forEach((pc) => {
                if (pc.state === types_1.PeerConnectionState.CONNECTING || pc.state === types_1.PeerConnectionState.NEW) {
                    this.processOnNotAnswer(pc);
                }
            });
            this.waitingOfferOrAnswerTimer = null;
        }, timeout * 1000);
    }
    toString() {
        return `ID: ${this.ID}, initiatorID: ${this.initiatorID}, opponentsIDs: ${this.opponentsIDs}, state: ${this.state}, callType: ${this.callType}`;
    }
}
exports.default = WebRTCSession;
function generateUUID() {
    let d = Date.now();
    const uuid = 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, (c) => {
        const r = (d + Math.random() * 16) % 16 | 0;
        d = Math.floor(d / 16);
        return (c == 'x' ? r : (r & 0x3) | 0x8).toString(16);
    });
    return uuid;
}
function prepareExtension(extension = {}) {
    const ext = { userInfo: extension };
    try {
        if (Utils_1.default.isObject(extension)) {
            ext.userInfo = Utils_1.default.cloneObject(extension, true);
        }
        else {
            WebRTCHelpers_1.default.traceWarning('Ignore "prepareExtension", must be an object');
        }
    }
    catch (err) {
        WebRTCHelpers_1.default.traceWarning(err.message);
    }
    return ext;
}
//# sourceMappingURL=WebRTCSession.js.map