"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const platform_1 = require("../platform");
const Utils_1 = __importDefault(require("../Utils"));
const JanusClient_1 = __importDefault(require("./JanusClient"));
const types_1 = require("../types");
class ConferenceSession {
    client;
    id = `${Math.random()}`;
    currentUserDisplayName;
    localStream;
    mediaParams = {};
    onParticipantJoinedListener;
    onParticipantLeftListener;
    onSlowLinkListener;
    onRemoteStreamListener;
    onRemoteTracksUpdatedListener;
    onRemoteConnectionStateChangedListener;
    onDataChannelOpenedListener;
    onDataChannelMessageListener;
    onSessionConnectionStateChangedListener;
    onErrorListener;
    constructor(token) {
        this.client = new JanusClient_1.default(token);
        this.client.on(types_1.JanusEvent.PARTICIPANT_JOINED, this.onParticipantJoined);
        this.client.on(types_1.JanusEvent.PARTICIPANT_LEFT, this.onParticipantLeft);
        this.client.on(types_1.JanusEvent.REMOTE_STREAM, this.onRemoteStream);
        this.client.on(types_1.JanusEvent.REMOTE_TRACKS_UPDATED, this.onRemoteTracksUpdated);
        this.client.on(types_1.JanusEvent.DATA_CHANNEL_OPEN, this.onDataChannelOpen);
        this.client.on(types_1.JanusEvent.DATA_CHANNEL_MESSAGE, this.onDataChannelMessage);
        this.client.on(types_1.JanusEvent.ERROR, this.onError);
    }
    get currentRoomId() {
        return this.client.currentRoomId;
    }
    set currentRoomId(roomId) {
        this.client.currentRoomId = roomId;
    }
    get currentPublisherPC() {
        return this.client.videoRoomPlugin.webrtcStuff.pc;
    }
    async createSession() {
        let isResolved = false;
        return new Promise((resolve, reject) => {
            this.client.createSession({
                success: () => {
                    isResolved = true;
                    resolve(void 0);
                },
                error: (error) => {
                    if (isResolved) {
                        this.onError(error);
                    }
                    else {
                        reject(error);
                    }
                },
            });
        });
    }
    // joinAsPublisher+Listener
    async join(roomId, user_id, userDisplayName) {
        this.currentUserDisplayName = userDisplayName;
        this.currentRoomId = roomId;
        await this.createSession();
        await this.createHandler(user_id, false, false);
        await this.joinInternal(this.currentRoomId, user_id);
    }
    async joinAsListener(roomId, user_id, userDisplayName) {
        this.currentUserDisplayName = userDisplayName;
        this.currentRoomId = roomId;
        await this.createSession();
        await this.createHandler(user_id, false, true);
        await this.joinInternal(this.currentRoomId, user_id);
    }
    async sendKeyframeRequest(roomId) {
        return new Promise((resolve, reject) => {
            this.client.sendKeyframeRequest(roomId, {
                success: resolve,
                error: reject,
            });
        });
    }
    async createHandler(user_id, isRemote, skipMedia = false) {
        return new Promise((resolve, reject) => {
            this.client.attachVideoConferencingPlugin(isRemote, user_id, skipMedia, {
                success: resolve,
                error: reject,
                iceState: isRemote
                    ? (iceState) => this.onRemoteIceStateChanged(user_id, iceState)
                    : (iceState) => this.onLocalIceStateChanged(iceState),
                slowLink: isRemote
                    ? (uplink, nacks) => this.onSlowLink(user_id, uplink, nacks)
                    : (uplink, nacks) => this.onSlowLink(null, uplink, nacks),
                localStream: this.localStream,
                displayName: this.currentUserDisplayName,
            });
        });
    }
    async joinInternal(roomId, user_id) {
        return new Promise((resolve, reject) => {
            this.client.join(roomId, user_id, false, {
                success: resolve,
                error: reject,
                displayName: this.currentUserDisplayName,
            });
        });
    }
    onParticipantJoined = (user_id, userDisplayName, isExistingParticipant) => {
        Utils_1.default.DLog('[onParticipantJoined]', user_id, userDisplayName, isExistingParticipant);
        this.createHandler(user_id, true, false);
        Utils_1.default.safeCallbackCall(this.onParticipantJoinedListener)(this, user_id, userDisplayName, isExistingParticipant);
    };
    onParticipantLeft = (user_id, userDisplayName) => {
        Utils_1.default.DLog('[onParticipantLeft]', user_id, userDisplayName);
        Utils_1.default.safeCallbackCall(this.onParticipantLeftListener)(this, user_id, userDisplayName);
    };
    onError = (error) => {
        Utils_1.default.DLog('[onError]', JSON.stringify(error));
        Utils_1.default.safeCallbackCall(this.onErrorListener)(this, error);
    };
    onDataChannelOpen = (label) => {
        Utils_1.default.DLog('[onDataChannelOpen]', label);
        Utils_1.default.safeCallbackCall(this.onDataChannelOpenedListener)(this, label);
    };
    onDataChannelMessage = (user_id, data) => {
        Utils_1.default.DLog('[onDataChannelMessage]', user_id, data);
        Utils_1.default.safeCallbackCall(this.onDataChannelMessageListener)(this, user_id, data);
    };
    onLocalIceStateChanged = (iceState) => {
        Utils_1.default.DLog('[onLocalIceStateChanged]', iceState);
        Utils_1.default.safeCallbackCall(this.onSessionConnectionStateChangedListener)(this, iceState);
    };
    onRemoteIceStateChanged = (user_id, iceState) => {
        Utils_1.default.DLog('[onRemoteIceStateChanged]', user_id, iceState);
        Utils_1.default.safeCallbackCall(this.onRemoteConnectionStateChangedListener)(this, user_id, iceState);
    };
    onRemoteStream = (user_id, stream) => {
        Utils_1.default.DLog('[onRemoteStream]', user_id, stream);
        Utils_1.default.safeCallbackCall(this.onRemoteStreamListener)(this, user_id, stream);
    };
    onRemoteTracksUpdated = (user_id, track, eventType) => {
        Utils_1.default.DLog('[onRemoteTracksUpdated]', user_id, track, eventType);
        Utils_1.default.safeCallbackCall(this.onRemoteTracksUpdatedListener)(this, user_id, track, eventType);
    };
    onSlowLink = (user_id, uplink, nacks) => {
        Utils_1.default.DLog('[onSlowLink]', user_id, uplink, nacks);
        Utils_1.default.safeCallbackCall(this.onSlowLinkListener)(this, user_id, uplink, nacks);
    };
    async listOfOnlineParticipants() {
        return new Promise((resolve, reject) => {
            if (this.currentRoomId) {
                this.client.listOnlineParticipants(this.currentRoomId, {
                    success: resolve,
                    error: reject,
                });
            }
            else {
                reject('Room ID is not set');
            }
        });
    }
    async leave() {
        this.currentRoomId = null;
        this.currentUserDisplayName = undefined;
        await this.leaveGroup();
        await this.detachVideoConferencingPlugin();
        await this.destroy();
        if (this.localStream) {
            this.localStream.getTracks().forEach((track) => track.stop());
            this.localStream = undefined;
            this.mediaParams = {};
        }
    }
    async leaveGroup() {
        return new Promise((resolve, reject) => {
            this.client.leave({
                success: resolve,
                error: reject,
            });
        });
    }
    async destroy() {
        return new Promise((resolve, reject) => {
            this.client.destroySession({
                success: resolve,
                error: reject,
            });
        });
    }
    async detachVideoConferencingPlugin() {
        return new Promise((resolve, reject) => {
            this.client.detachVideoConferencingPlugin({
                success: resolve,
                error: reject,
            });
        });
    }
    async getDisplayMedia(params) {
        if (!platform_1.mediaDevices.getDisplayMedia) {
            throw new Error("Your environment does not support 'getDisplayMedia' API");
        }
        const elementId = params && params.elementId;
        const options = params && params.options;
        const mediaParams = { ...params };
        this.mediaParams = mediaParams;
        delete mediaParams.elementId;
        delete mediaParams.options;
        try {
            const stream = await platform_1.mediaDevices.getDisplayMedia(mediaParams);
            return this.upsertStream(stream, elementId, options);
        }
        catch (error) {
            throw error;
        }
    }
    async getUserMedia(params) {
        const elementId = params && params.elementId;
        const options = params && params.options;
        const mediaParams = { ...params };
        this.mediaParams = mediaParams;
        delete mediaParams.elementId;
        delete mediaParams.options;
        try {
            const stream = await platform_1.mediaDevices.getUserMedia(mediaParams);
            return this.upsertStream(stream, elementId, options);
        }
        catch (error) {
            throw error;
        }
    }
    upsertStream(stream, elementId = null, options = {}) {
        const shouldUpdateCurrentStream = !!this.localStream;
        if (shouldUpdateCurrentStream) {
            this.replaceTracks(stream);
        }
        else {
            this.localStream = stream;
        }
        if (!this.localStream) {
            throw new Error('Local stream is not defined');
        }
        if (elementId) {
            if (shouldUpdateCurrentStream) {
                this.detachMediaStream(elementId, options);
            }
            this.attachMediaStream(elementId, this.localStream, options);
        }
        return this.localStream;
    }
    replaceTracks(stream) {
        this.localStream?.getTracks().forEach((localTrack) => {
            if (localTrack.kind === types_1.JanusCallType.AUDIO && stream.getAudioTracks().length === 0) {
                return;
            }
            else {
                localTrack.stop();
                this.localStream?.removeTrack(localTrack);
            }
        });
        stream.getTracks().forEach((newTrack) => {
            const pcSenders = this.currentPublisherPC.getSenders();
            const sender = pcSenders.find(({ track }) => newTrack.kind === track.kind);
            if (newTrack.kind === types_1.JanusCallType.AUDIO) {
                newTrack.enabled = this.localStream?.getAudioTracks().every(({ enabled }) => enabled) ?? true;
            }
            else {
                newTrack.enabled = this.localStream?.getVideoTracks().every(({ enabled }) => enabled) ?? true;
            }
            if (sender) {
                sender.replaceTrack(newTrack);
                this.localStream?.addTrack(newTrack);
            }
            else {
                console.warn(`No sender found for track kind: ${newTrack.kind}`);
            }
        });
        if (!this.localStream) {
            throw new Error('Local stream is not defined');
        }
        return this.localStream;
    }
    async switchMediaTracks(options) {
        [types_1.JanusCallType.AUDIO, types_1.JanusCallType.VIDEO].forEach((type) => {
            const device = type === types_1.JanusCallType.AUDIO ? options.audio : type === types_1.JanusCallType.VIDEO ? options.video : {};
            const deviceId = typeof device === 'string' ? device : device?.deviceId ?? null;
            if (deviceId) {
                if (typeof this.mediaParams[type] === 'object') {
                    this.mediaParams[type].deviceId = deviceId;
                }
                else {
                    this.mediaParams[type] = { deviceId };
                }
            }
        });
        try {
            const mediaParams = {
                audio: this.mediaParams?.audio ?? false,
                video: this.mediaParams?.video ?? false,
            };
            const stream = await platform_1.mediaDevices.getDisplayMedia(mediaParams);
            return this.replaceTracks(stream);
        }
        catch (error) {
            throw error;
        }
    }
    muteVideo() {
        if (!this.isVideoMuted()) {
            this.client.toggleVideoMute();
        }
    }
    unmuteVideo() {
        if (this.isVideoMuted()) {
            this.client.toggleVideoMute();
        }
    }
    muteAudio() {
        if (!this.isAudioMuted()) {
            this.client.toggleAudioMute();
        }
    }
    unmuteAudio() {
        if (this.isAudioMuted()) {
            this.client.toggleAudioMute();
        }
    }
    isVideoMuted() {
        return this.client.isVideoMuted();
    }
    isAudioMuted() {
        return this.client.isAudioMuted();
    }
    async getUserVolume() {
        return new Promise((resolve, _reject) => {
            return this.client.getVolume(resolve);
        });
    }
    getRemoteUserBitrate(userId) {
        return this.client.getUserBitrate(userId);
    }
    async getRemoteUserVolume(userId) {
        return new Promise((resolve, reject) => {
            return this.client.getUserVolume(userId, resolve);
        });
    }
    attachMediaStream(elementId, stream, opt) {
        const mediaElement = document.getElementById(elementId);
        if ('srcObject' in mediaElement) {
            mediaElement.srcObject = stream;
            mediaElement.style.transform = opt?.mirror ? 'scaleX(-1)' : 'none';
            if (typeof opt?.muted === 'boolean') {
                mediaElement.muted = opt?.muted;
            }
            mediaElement.onloadedmetadata = (_event) => {
                mediaElement.play();
            };
        }
        else {
            throw new Error(`Unable to attach media stream, element #${elementId} is undefined`);
        }
    }
    detachMediaStream(elementId, opt) {
        const mediaElement = document.getElementById(elementId);
        if ('srcObject' in mediaElement) {
            mediaElement.pause();
            mediaElement.srcObject = null;
            mediaElement.style.transform = opt?.mirror ? 'scaleX(-1)' : 'none';
            if (typeof opt?.muted === 'boolean') {
                mediaElement.muted = opt?.muted;
            }
        }
        else {
            throw new Error(`Unable to attach media stream, element #${elementId} is undefined`);
        }
    }
    async sendData(data, label) {
        return new Promise((resolve, reject) => {
            this.client.videoRoomPlugin.data({
                data,
                label,
                success: resolve,
                error: reject,
            });
        });
    }
}
exports.default = ConferenceSession;
//# sourceMappingURL=ConferenceSession.js.map