"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const eventemitter3_1 = __importDefault(require("eventemitter3"));
const platform_1 = require("../platform");
const config_1 = __importDefault(require("../config"));
const janus_1 = __importDefault(require("./janus"));
const Utils_1 = __importDefault(require("../Utils"));
const types_1 = require("../types");
class JanusClient {
    token;
    server;
    debug;
    engine;
    videoRoomPlugin = null;
    isOnlyAudio = false;
    currentRoomId = null;
    currentUserId = null;
    remoteFeeds = {};
    remoteJseps = {};
    remoteFeedsAttachingInProgress = {};
    bitrateTimers = {};
    emitter = new eventemitter3_1.default();
    constructor(token) {
        if (!Utils_1.default.env.isReactNative && !platform_1.adapter) {
            throw 'Error: in order to use this library please connect adapter.js. More info https://github.com/webrtc/adapter';
        }
        this.token = token;
        this.server = config_1.default.conference.server;
        this.debug = config_1.default.conference.debug ?? types_1.JanusDebugMode.ALL;
        if (!this.server) {
            throw "'server' parameter is mandatory";
        }
        else if (this.server.includes('http')) {
            this.server += '/janus';
        }
    }
    createSession({ success = () => { }, error = () => { }, destroyed = () => { }, timeoutSessionCallback = () => { }, }) {
        janus_1.default.init({
            debug: this.debug,
            callback: () => {
                if (!janus_1.default.isWebrtcSupported()) {
                    Utils_1.default.safeCallbackCall(error)(`Your browser does not support WebRTC, so you can't use this functionality.`);
                    return;
                }
                this.engine = new janus_1.default({
                    server: this.server,
                    iceServers: config_1.default.videochat.iceServers,
                    token: this.token,
                    success: () => Utils_1.default.safeCallbackCall(success)(),
                    error: (err) => Utils_1.default.safeCallbackCall(error)(err),
                    destroyed: () => Utils_1.default.safeCallbackCall(destroyed)(),
                    timeoutSessionCallback: () => Utils_1.default.safeCallbackCall(timeoutSessionCallback)(),
                });
            },
        });
    }
    getSessionId() {
        return this.engine?.getSessionId() ?? null;
    }
    destroySession({ success = () => { }, error = () => { } }) {
        this.engine.destroy({
            success: () => Utils_1.default.safeCallbackCall(success)(),
            error: (err) => Utils_1.default.safeCallbackCall(error)(err),
        });
    }
    attachVideoConferencingPlugin(isRemote, userId, skipMedia, callbacks) {
        let remoteFeed = null;
        const localStream = callbacks.localStream;
        delete callbacks.localStream;
        const displayName = callbacks.displayName;
        delete callbacks.displayName;
        this.engine.attach({
            plugin: 'janus.plugin.videoroom',
            success: (pluginHandle) => {
                if (isRemote) {
                    remoteFeed = pluginHandle;
                    remoteFeed.userId = userId;
                    this.remoteFeedsAttachingInProgress[userId] = remoteFeed;
                    const listen = {
                        request: 'join',
                        room: this.currentRoomId,
                        ptype: 'listener',
                        feed: userId,
                        display: displayName,
                    };
                    remoteFeed?.send({ message: listen });
                }
                else {
                    this.videoRoomPlugin = pluginHandle;
                }
                Utils_1.default.safeCallbackCall(callbacks.success)();
            },
            error: (error) => {
                Utils_1.default.safeCallbackCall(callbacks.error)(error);
            },
            consentDialog: (on) => {
                Utils_1.default.safeCallbackCall(callbacks.consentDialog)(on);
            },
            mediaState: (medium, on) => {
                Utils_1.default.safeCallbackCall(callbacks.mediaState)(medium, on);
            },
            webrtcState: (on) => {
                Utils_1.default.safeCallbackCall(callbacks.webrtcState)(on);
            },
            slowLink: (uplink, nacks) => {
                Utils_1.default.safeCallbackCall(callbacks.slowLink)(uplink, nacks);
            },
            iceState: (iceConnectionState) => {
                Utils_1.default.safeCallbackCall(callbacks.iceState)(iceConnectionState);
            },
            onmessage: (msg, jsep) => {
                const event = msg['videoroom'];
                if (isRemote) {
                    if (event) {
                        // Remote feed attached
                        if (event === 'attached') {
                            const feedId = msg['id'];
                            this.remoteFeeds[feedId] = this.remoteFeedsAttachingInProgress[feedId];
                            this.remoteFeedsAttachingInProgress[feedId] = null;
                        }
                        else if (msg['error']) {
                            Utils_1.default.safeCallbackCall(callbacks.error)(msg['error']);
                        }
                    }
                    if (jsep) {
                        const feedId = msg['id'];
                        // ICE restart case
                        if (!feedId) {
                        }
                        this.remoteJseps[feedId] = jsep;
                        this.createAnswer({
                            remoteFeed: this.remoteFeeds[feedId],
                            jsep,
                        }, localStream, {
                            success: () => { },
                            error: (error) => {
                                Utils_1.default.safeCallbackCall(callbacks.error)(error);
                            },
                        });
                    }
                    // local feed
                }
                else {
                    if (event) {
                        // We JOINED
                        if (event === 'joined') {
                            const media = skipMedia ? { audio: false, video: false } : { audio: true, video: true };
                            const existedStream = skipMedia ? undefined : localStream;
                            const offerParams = { media, stream: existedStream };
                            this.createOffer(offerParams, {
                                success: () => {
                                    if (msg['publishers']) {
                                        const publishers = msg['publishers'];
                                        for (const f in publishers) {
                                            const userId = publishers[f]['id'];
                                            const userDisplayName = publishers[f]['display'];
                                            this.emitter.emit(types_1.JanusEvent.PARTICIPANT_JOINED, userId, userDisplayName, true);
                                        }
                                    }
                                },
                                error: (error) => {
                                    Utils_1.default.safeCallbackCall(callbacks.error)(error);
                                },
                            });
                            // We JOINED and now receiving who is online
                        }
                        else if (event === 'event') {
                            // Any new feed to attach to?
                            if (msg['publishers']) {
                                const publishers = msg['publishers'];
                                for (const f in publishers) {
                                    const userId = publishers[f]['id'];
                                    const userDisplayName = publishers[f]['display'];
                                    this.emitter.emit(types_1.JanusEvent.PARTICIPANT_JOINED, userId, userDisplayName, false);
                                }
                                // Someone is LEAVING
                            }
                            else if (msg['leaving']) {
                                // One of the publishers has gone away?
                                const feedId = msg['leaving'];
                                const success = this.detachRemoteFeed(feedId);
                                if (success) {
                                    this.emitter.emit(types_1.JanusEvent.PARTICIPANT_LEFT, feedId, null);
                                }
                            }
                            else if (msg['unpublished']) {
                                // One of the publishers has gone away?
                                const feedId = msg['unpublished'];
                                if (feedId != 'ok') {
                                    const success = this.detachRemoteFeed(feedId);
                                    if (success) {
                                        this.emitter.emit(types_1.JanusEvent.PARTICIPANT_LEFT, feedId, null);
                                    }
                                }
                            }
                            else if (msg['error']) {
                                Utils_1.default.DLog('[janus error message]', msg['error']);
                                // #define VIDEOROOM_ERROR_ID_EXISTS			436
                                // #define VIDEOROOM_ERROR_UNAUTHORIZED		433
                                //
                                this.emitter.emit(types_1.JanusEvent.ERROR, msg);
                                Utils_1.default.safeCallbackCall(callbacks.error)(msg['error']);
                            }
                        }
                    }
                    if (jsep) {
                        this.videoRoomPlugin.handleRemoteJsep({ jsep: jsep });
                        // TODO:
                        // handle wrong or unsupported codecs here...
                        // const video = msg['video_codec']
                        // if(mystream && mystream.getVideoTracks() && mystream.getVideoTracks().length > 0 && !video) {
                        // 		'Our video stream has been rejected, viewers won't see us'
                        // }
                    }
                }
            },
            onlocaltrack: (track, on) => {
                Utils_1.default.DLog('[onlocaltrack]', track, on);
                this.onLocalTrack(track, on);
            },
            onremotetrack: (track, mid, on, metadata) => {
                Utils_1.default.DLog('[onremotetrack]', track, mid, on, metadata);
                this.onRemoteTrack(remoteFeed, track, mid, on, metadata);
            },
            ondataopen: (channelLabel) => {
                Utils_1.default.DLog('[ondataopen]', channelLabel);
                this.emitter.emit(types_1.JanusEvent.DATA_CHANNEL_OPEN, channelLabel);
            },
            ondata: (data, channelLabel) => {
                Utils_1.default.DLog('[ondata]', channelLabel, data);
                this.emitter.emit(types_1.JanusEvent.DATA_CHANNEL_MESSAGE, channelLabel, data);
            },
            oncleanup: () => {
                Utils_1.default.safeCallbackCall(callbacks.oncleanup)();
            },
            detached: () => { },
        });
    }
    onLocalTrack(track, on) {
        // this.emitter.emit(JanusEvent.LOCAL_STREAM, track);
    }
    onRemoteTrack(remoteFeed, track, mid, on, metadata) {
        const eventType = metadata && metadata.reason;
        if (eventType === types_1.JanusMediaTrackReason.CREATED) {
            const isStreamNoExistedBefore = !remoteFeed.stream || !remoteFeed.tracks;
            if (isStreamNoExistedBefore) {
                remoteFeed.tracks = { [mid]: track };
                remoteFeed.stream = new platform_1.MediaStream([track]);
            }
            else {
                remoteFeed.tracks[mid] = track;
                remoteFeed.stream.addTrack(track);
            }
            if (isStreamNoExistedBefore) {
                this.emitter.emit(types_1.JanusEvent.REMOTE_STREAM, remoteFeed.userId, remoteFeed.stream);
            }
        }
        else if (eventType === types_1.JanusMediaTrackReason.ENDED) {
            delete remoteFeed.tracks[mid];
            const trackToRemove = remoteFeed.stream.getTracks().find((streamTrack) => streamTrack.kind === track.kind);
            remoteFeed.stream.removeTrack(trackToRemove);
        }
        this.emitter.emit(types_1.JanusEvent.REMOTE_TRACKS_UPDATED, remoteFeed.userId, track, eventType);
    }
    getPluginId() {
        return this.videoRoomPlugin?.getId() ?? null;
    }
    detachVideoConferencingPlugin(callbacks) {
        const clean = () => {
            this.videoRoomPlugin = null;
            // detach all remote feeds
            Object.keys(this.remoteFeeds).forEach((userId) => {
                this.detachRemoteFeed(Number(userId));
            });
            this.remoteFeeds = {};
            this.remoteJseps = {};
            /*
             * Deprecated
            this.currentMediaDeviceId = null
             */
        };
        this.videoRoomPlugin.detach({
            success: () => {
                clean();
                Utils_1.default.safeCallbackCall(callbacks.success)();
            },
            error: (error) => {
                clean();
                Utils_1.default.safeCallbackCall(callbacks.error)(error);
            },
        });
    }
    join(roomId, userId, isOnlyAudio, callbacks) {
        const displayName = callbacks.displayName;
        delete callbacks.displayName;
        this.isOnlyAudio = isOnlyAudio;
        Utils_1.default.DLog('isOnlyAudio: ' + this.isOnlyAudio);
        const joinEvent = {
            request: 'join',
            room: roomId,
            ptype: 'publisher',
            id: userId,
            display: displayName,
        };
        this.videoRoomPlugin.send({
            message: joinEvent,
            success: (resp) => {
                this.currentRoomId = roomId;
                this.currentUserId = userId;
                Utils_1.default.safeCallbackCall(callbacks.success)();
            },
            error: (error) => {
                Utils_1.default.safeCallbackCall(callbacks.error)(error);
            },
        });
    }
    leave(callbacks) {
        Utils_1.default.DLog('leave');
        if (!this.engine.isConnected()) {
            Utils_1.default.safeCallbackCall(callbacks.success)();
            return;
        }
        const leaveEvent = {
            request: 'leave',
            room: this.currentRoomId,
            id: this.currentUserId,
        };
        if (this.videoRoomPlugin) {
            this.videoRoomPlugin.send({ message: leaveEvent });
        }
        this.currentRoomId = null;
        this.currentUserId = null;
        Utils_1.default.safeCallbackCall(callbacks.success)();
    }
    listOnlineParticipants(roomId, callbacks) {
        const listRequest = { request: 'listparticipants', room: roomId };
        this.videoRoomPlugin.send({
            message: listRequest,
            success: (data) => {
                const participants = data ? data.participants : [];
                Utils_1.default.safeCallbackCall(callbacks.success)(participants);
            },
            error: (error) => {
                Utils_1.default.safeCallbackCall(callbacks.error)(error);
            },
        });
    }
    toggleAudioMute() {
        const muted = this.videoRoomPlugin.isAudioMuted();
        if (muted) {
            this.videoRoomPlugin.unmuteAudio();
        }
        else {
            this.videoRoomPlugin.muteAudio();
        }
        return this.videoRoomPlugin.isAudioMuted();
    }
    isAudioMuted() {
        return this.videoRoomPlugin.isAudioMuted();
    }
    toggleRemoteAudioMute(userId) {
        const remoteFeed = this.remoteFeeds[userId];
        if (!remoteFeed) {
            return false;
        }
        const audioTracks = remoteFeed.stream.getAudioTracks();
        if (audioTracks && audioTracks.length > 0) {
            for (let i = 0; i < audioTracks.length; ++i) {
                audioTracks[i].enabled = !audioTracks[i].enabled;
            }
            return !audioTracks[0].enabled;
        }
        return false;
    }
    isRemoteAudioMuted(userId) {
        const remoteFeed = this.remoteFeeds[userId];
        if (!remoteFeed) {
            return false;
        }
        const audioTracks = remoteFeed.stream.getAudioTracks();
        if (audioTracks && audioTracks.length > 0) {
            return !audioTracks[0].enabled;
        }
        return false;
    }
    toggleVideoMute() {
        const muted = this.videoRoomPlugin.isVideoMuted();
        if (muted) {
            this.videoRoomPlugin.unmuteVideo();
        }
        else {
            this.videoRoomPlugin.muteVideo();
        }
        return this.videoRoomPlugin.isVideoMuted();
    }
    isVideoMuted() {
        return this.videoRoomPlugin.isVideoMuted();
    }
    toggleRemoteVideoMute(userId) {
        const remoteFeed = this.remoteFeeds[userId];
        if (!remoteFeed) {
            return false;
        }
        const videoTracks = remoteFeed.stream.getVideoTracks();
        if (videoTracks && videoTracks.length > 0) {
            for (let i = 0; i < videoTracks.length; ++i) {
                videoTracks[i].enabled = !videoTracks[i].enabled;
            }
            return !videoTracks[0].enabled;
        }
        return false;
    }
    isRemoteVideoMuted(userId) {
        const remoteFeed = this.remoteFeeds[userId];
        if (!remoteFeed) {
            return false;
        }
        const videoTracks = remoteFeed.stream.getVideoTracks();
        if (videoTracks && videoTracks.length > 0) {
            return !videoTracks[0].enabled;
        }
        return false;
    }
    sendKeyframeRequest(roomId, callbacks) {
        const configureRequest = {
            request: 'configure',
            room: roomId,
            keyframe: true,
        };
        this.videoRoomPlugin.send({
            message: configureRequest,
            success: (response) => {
                Utils_1.default.safeCallbackCall(callbacks.success)(response);
            },
            error: (error) => {
                Utils_1.default.safeCallbackCall(callbacks.error)(error);
            },
        });
    }
    getTracksFromStream(stream) {
        const tracks = [];
        const audioTracks = stream.getAudioTracks();
        if (audioTracks.length) {
            const audioTrack = audioTracks[0];
            tracks.push({ type: 'audio', capture: audioTrack, recv: false });
        }
        const videoTracks = stream.getVideoTracks();
        if (videoTracks.length) {
            const videoTrack = videoTracks[0];
            tracks.push({ type: 'video', capture: videoTrack, recv: false });
        }
        return tracks;
    }
    createOffer(mediaParams, callbacks) {
        Utils_1.default.DLog('[JanusWrapper][createOffer]', mediaParams);
        const { stream: existedStream, media, replace } = mediaParams;
        const createOfferParams = { tracks: [{ type: 'data' }] };
        if (existedStream) {
            const tracksFromStream = this.getTracksFromStream(existedStream);
            createOfferParams.tracks = createOfferParams.tracks.concat(tracksFromStream);
        }
        else if (media) {
            const tracksFromParams = [];
            if (media.audio) {
                tracksFromParams.push({
                    type: 'audio',
                    capture: media.audio,
                    recv: false,
                    replace: !!replace,
                });
            }
            if (media.video) {
                tracksFromParams.push({
                    type: 'video',
                    capture: media.video,
                    recv: false,
                    replace: !!replace,
                });
            }
            createOfferParams.tracks = createOfferParams.tracks.concat(tracksFromParams);
        }
        else {
            createOfferParams.tracks = createOfferParams.tracks.concat([
                { type: 'audio', capture: true, recv: false, replace: !!replace },
                { type: 'video', capture: true, recv: false, replace: !!replace },
            ]);
        }
        Utils_1.default.DLog('[JanusWrapper][createOffer][params]', createOfferParams);
        createOfferParams.customizeSdp = (jsep) => { };
        createOfferParams.success = (jsep) => {
            const publish = {
                request: 'configure',
                audio: !!media.audio,
                video: !!media.video,
            };
            Utils_1.default.DLog('[JanusWrapper][createOffer][success]', publish);
            this.videoRoomPlugin.send({ message: publish, jsep: jsep });
            Utils_1.default.safeCallbackCall(callbacks.success)();
        };
        createOfferParams.error = (error) => {
            Utils_1.default.DLog('[JanusWrapper][createOffer][error]', error);
            if (media.audio) {
                this.createOffer({ media: { video: false, audio: false } }, callbacks);
            }
            else {
                Utils_1.default.safeCallbackCall(callbacks.error)(error);
            }
        };
        this.videoRoomPlugin.createOffer(createOfferParams);
    }
    getTracksMidsFromStream(stream) {
        const tracks = [];
        const audioTracks = stream.getAudioTracks();
        if (audioTracks.length) {
            const audioTrack = audioTracks[0];
            tracks.push({ type: 'audio', mid: audioTrack.id, recv: true });
        }
        const videoTracks = stream.getVideoTracks();
        if (videoTracks.length) {
            const videoTrack = videoTracks[0];
            tracks.push({ type: 'video', mid: videoTrack.id, recv: true });
        }
        return tracks;
    }
    createAnswer({ remoteFeed, jsep }, existedStream, callbacks) {
        Utils_1.default.DLog('[JanusWrapper][createAnswer]', jsep, existedStream);
        let tracks = [{ type: 'data' }];
        if (existedStream) {
            const tracksFromStream = this.getTracksMidsFromStream(existedStream);
            tracks = tracks.concat(tracksFromStream);
        }
        Utils_1.default.DLog('[JanusWrapper][createAnswer][tracks]', tracks);
        remoteFeed.createAnswer({
            jsep: jsep,
            tracks: tracks,
            success: (jsep) => {
                const body = { request: 'start', room: this.currentRoomId };
                Utils_1.default.DLog('[JanusWrapper][createAnswer][success]', body);
                remoteFeed.send({ message: body, jsep: jsep });
                Utils_1.default.safeCallbackCall(callbacks.success)();
            },
            error: (error) => {
                Utils_1.default.DLog('[JanusWrapper][createAnswer][error]', error);
                Utils_1.default.safeCallbackCall(callbacks.error)(error);
            },
        });
    }
    detachRemoteFeed(userId) {
        const remoteFeed = this.remoteFeeds[userId];
        if (remoteFeed) {
            remoteFeed.detach();
            this.remoteFeeds[userId] = null;
            this.remoteJseps[userId] = null;
            return true;
        }
        return false;
    }
    getUserBitrate(userId) {
        const remoteFeed = this.remoteFeeds[userId];
        return remoteFeed.getBitrate();
    }
    getVolume(resultCallback) {
        return this.videoRoomPlugin.getLocalVolume(null, resultCallback);
    }
    getUserVolume(userId, resultCallback) {
        const remoteFeed = this.remoteFeeds[userId];
        return remoteFeed.getRemoteVolume(null, resultCallback);
    }
    showBitrate(userId, element) {
        const remoteFeed = this.remoteFeeds[userId];
        if (!Utils_1.default.env.isReactNative &&
            (platform_1.adapter.browserDetails.browser === 'chrome' || platform_1.adapter.browserDetails.browser === 'firefox')) {
            this.bitrateTimers[userId] = setInterval(() => {
                const bitrate = remoteFeed.getBitrate();
                element.text(bitrate);
            }, 1000);
        }
    }
    hideBitrate(userId, element) {
        if (this.bitrateTimers[userId]) {
            clearInterval(this.bitrateTimers[userId]);
        }
        this.bitrateTimers[userId] = null;
        element.text = null;
    }
    on(eventType, listener) {
        return this.emitter.addListener(eventType, listener);
    }
    removeAllListeners(eventType) {
        return this.emitter.removeAllListeners(eventType);
    }
}
exports.default = JanusClient;
//# sourceMappingURL=JanusClient.js.map